/*
 * Romeo_Wrap.cpp : Wrapper between Move3D and HPP that provides an interface for handling Romeo
 *
 * Author : renaud Viry (renaud.viry@laas.fr)
 * Copyright LAAS-CNRS 2014
 *
 */
#include "Humanoid_Wrap.hpp"
#include <math.h>

#ifdef HPP
#include <hpp/model/collision-object.hh>

/*********************************************************************************************************************
 * Global functions
 *********************************************************************************************************************/
/* Create stability constraints
 *
 * \param robot is the sharedPtr to teh device the constraints is defined for (i.e. the Humanoid Robot)
 * \param leftAnkle is teh sharedPtr to the joint of the left ankle of the robot
 * \param rightAnkle is teh sharedPtr to the joint of the right ankle of the robot
 * \param configuration is the input configuration
 * \return the vector of constraints for stability
 */
std::vector<hpp::core::DifferentiableFunctionPtr_t> createSlidingStabilityConstraint(const hpp::model::DevicePtr_t& robot,
																					 const hpp::model::JointPtr_t& leftAnkle,
																					 const hpp::model::JointPtr_t& rightAnkle,
																					 hpp::core::ConfigurationIn_t configuration)
{
	std::vector<hpp::core::DifferentiableFunctionPtr_t> result;

	robot->currentConfiguration (configuration);
	robot->computeForwardKinematics();

	hpp::model::JointPtr_t joint1 = leftAnkle;
	hpp::model::JointPtr_t joint2 = rightAnkle;

	const hpp::model::Transform3f& M1 = joint1->currentTransformation ();
	const hpp::model::Transform3f& M2 = joint2->currentTransformation ();
	const hpp::model::vector3_t& x = robot->positionCenterOfMass ();

	// position of center of mass in left ankle frame
	hpp::model::matrix3_t R1T (M1.getRotation ()); R1T.transpose ();
	hpp::model::vector3_t xloc = R1T * (x - M1.getTranslation ());
	result.push_back(hpp::constraints::RelativeCom::create(robot, joint1, xloc));

	// Relative orientation of the feet
	hpp::model::matrix3_t reference = R1T * M2.getRotation ();
	result.push_back(hpp::constraints::RelativeOrientation::create(robot, joint1, joint2, reference));

	// Relative position of the feet
	hpp::model::vector3_t local1; local1.setZero ();
	hpp::model::vector3_t global1 = M1.getTranslation ();
	hpp::model::matrix3_t R2T (M2.getRotation ()); R2T.transpose ();
	hpp::model::vector3_t local2 = R2T * (global1 - M2.getTranslation ());
	result.push_back (hpp::constraints::RelativePosition::create(robot, joint1, joint2, local1, local2));

	// Orientation of the left foot
	reference.setIdentity ();
	result.push_back(hpp::constraints::Orientation::create (robot, joint1, reference,
										  boost::assign::list_of(true)(true)(false)));

	// Position of the left foot
	hpp::model::vector3_t zero; zero.setZero ();
	hpp::model::matrix3_t I3; I3.setIdentity ();
	result.push_back (hpp::constraints::Position::create(robot, joint1, zero, M1.getTranslation (), I3,
									   boost::assign::list_of (false)(false)(true)));

	return result;
}

/* Convert a sequence of DoF value to a robot configuration
 *
 * \param ps is the problem solver where the constraints are stored
 * \param dofArray is the input sequence of DoF values
 * \return a sharePtr to the corresponding configuration
 */
static hpp::core::ConfigurationPtr_t dofSeqToConfig(hpp::core::ProblemSolverPtr_t ps,
													const std::vector<double> dofArray)
{
	unsigned int configDim = (unsigned int)dofArray.size();
	hpp::core::ConfigurationPtr_t config (new hpp::core::Configuration_t (configDim));

	// Get robot in hppPlanner object.
	hpp::model::DevicePtr_t robot = ps->robot ();

	// Compare size of input array with number of degrees of freedom of
	if (configDim != robot->configSize ())
	{
		std::cout << "ERROR : robot configSize (" << robot->configSize ()
				  << ") is different from config size ("
				  << configDim << ")" << std::endl;
		throw std::runtime_error("robot nb dof is different from config size");
	}

	// Fill dof vector with dof array.
	for (unsigned int iDof=0; iDof < configDim; ++iDof)
		(*config) [iDof] = dofArray [iDof];

	return config;
}

/*********************************************************************************************************************/

HumanoidWrap::HumanoidWrap()
{
}

bool HumanoidWrap::loadHumanoidRobot(const std::string& robotName, const std::string& rootJointType,
									 const std::string& packageName, const std::string& modelName,
									 const std::string& urdfSuffix, const std::string& srdfSuffix,
									 Robot *move3dRobot)
{
	_move3dRobot = move3dRobot;
	hpp::model::HumanoidRobotPtr_t robot = hpp::model::HumanoidRobot::create (robotName);
	if (!robot)
	{
		std::cout << "ERROR : creating HumanoidRobot failed" << std::endl;
		return false;
	}

	hpp::model::urdf::loadHumanoidModel (robot, rootJointType, packageName, modelName, urdfSuffix, srdfSuffix);
	_manipulationSolver = new hpp::core::ManipulationSolver();
	if (!_manipulationSolver)
	{
		std::cout << "ERROR : loading HumanoidRobot model failed" << std::endl;
		return false;
	}

	_manipulationSolver->robot(robot);

	createStabilityConstraints();

	return true;
}

void HumanoidWrap::createStabilityConstraints()
{
	_numericalConstraints.clear();

	// Create configuration from the startConf map
	std::vector<double> dofArray;

	// Get the half-siting configuration
	hpp::model::JointVector_t robotJointVector = _manipulationSolver->robot()->getJointVector();
	for (int i = 0; i < robotJointVector.size(); i++)
	{
		if (robotJointVector[i]->name() == "base_joint_xyz")
		{
			dofArray.push_back(0);
			dofArray.push_back(0);
			dofArray.push_back(0.840252);
		}
		else if (robotJointVector[i]->name() == "base_joint_SO3")
		{
			dofArray.push_back(1);
			dofArray.push_back(0);
			dofArray.push_back(0);
			dofArray.push_back(0);
		}
		else
		{
			for (std::map<std::string, double>::iterator it = startConf.begin(); it!= startConf.end(); ++it)
			{
				if (robotJointVector[i]->name() == it->first)
				{
					// std::cout << "RV -- (" << i << ") Name of the joint : " << it->first << std::endl
					// 		  << "         Value : " << it->second << std::endl;
					dofArray.push_back(it->second);
					continue;
				}
			}
		}
	}

	// Create static stability constraint
	try
	{
		hpp::core::ConfigurationPtr_t config = dofSeqToConfig (_manipulationSolver, dofArray);

		if (!_manipulationSolver->robot())
		{
			throw std::runtime_error ("You should set the robot before defining constraints.");
		}
		
		_manipulationSolver->robot()->currentConfiguration(*config);

		hpp::model::JointPtr_t la = _manipulationSolver->robot()->getJointByName(_leftAnkle);
		hpp::model::JointPtr_t ra = _manipulationSolver->robot()->getJointByName(_rightAnkle);
		_numericalConstraints = createSlidingStabilityConstraint(_manipulationSolver->robot(), la, ra, *config);
	} catch (const std::exception& exc)
	{
		throw std::runtime_error (exc.what ());
	}

	// Add stability constraints to problemSolver and configProjector
	std::string p ("balance");
	_manipulationSolver->addNumericalConstraint(p + std::string ("/relative-com"), _numericalConstraints [0]);
	_manipulationSolver->addNumericalConstraint(p + std::string ("/relative-orientation"), _numericalConstraints [1]);
	_manipulationSolver->addNumericalConstraint(p + std::string ("/relative-position"), _numericalConstraints [2]);
	_manipulationSolver->addNumericalConstraint(p + std::string ("/orientation-left-foot"), _numericalConstraints [3]);
	_manipulationSolver->addNumericalConstraint(p + std::string ("/position-left-foot"), _numericalConstraints [4]);
	_manipulationSolver->addConstraintToConfigProjector(p, _manipulationSolver->numericalConstraint(p +
													std::string ("/relative-com")));
	_manipulationSolver->addConstraintToConfigProjector(p, _manipulationSolver->numericalConstraint(p +
													std::string ("/relative-orientation")));
	_manipulationSolver->addConstraintToConfigProjector(p, _manipulationSolver->numericalConstraint(p +
													std::string ("/relative-position")));
	_manipulationSolver->addConstraintToConfigProjector(p, _manipulationSolver->numericalConstraint(p +
													std::string ("/orientation-left-foot")));
	_manipulationSolver->addConstraintToConfigProjector(p, _manipulationSolver->numericalConstraint(p +
													std::string ("/position-left-foot")));
}

bool HumanoidWrap::isPositionStable(confPtr_t config)
{
	// Checks if the projection of COM in (X, Y) plane is within a bounded distance to the center-center of joints segment
	hpp::core::ConfigurationPtr_t configuration = createHppConfiguration(config);

	_manipulationSolver->robot()->currentConfiguration (*configuration);
	_manipulationSolver->robot()->computeForwardKinematics();

	const hpp::model::vector3_t& COM = _manipulationSolver->robot()->positionCenterOfMass ();

	p3d_vector3 t1, t2, com, vecRes, vecT1T2, vecT1com, vecT2T1, vecT2com;
	t1[0] = _move3dRobot->getJoint(_leftAnkle)->getJointStruct()->abs_pos[0][3];
	t1[1] = _move3dRobot->getJoint(_leftAnkle)->getJointStruct()->abs_pos[1][3];
	t1[2] = _move3dRobot->getJoint(_leftAnkle)->getJointStruct()->abs_pos[2][3];
	t2[0] = _move3dRobot->getJoint(_rightAnkle)->getJointStruct()->abs_pos[0][3];
	t2[1] = _move3dRobot->getJoint(_rightAnkle)->getJointStruct()->abs_pos[1][3];
	t2[2] = 0;
	com[0] = COM[0]; com[1] = COM[1]; com[2] = 0;

	if (t1[2] > 0.1)
	{
		std::cout << "ERROR : the left Ankle is above 5 cm from the ground (Z = " << t1[2] << ")" << std::endl;
		return false;
	}

	t1[2] = 0;	// Projection on (X, Y) plane

	p3d_vectSub(t2, t1, vecT1T2);
	p3d_vectSub(com, t1, vecT1com);
	p3d_vectSub(t1, t2, vecT2T1);
	p3d_vectSub(com, t2, vecT2com);

	double T1Com = p3d_vectNorm(vecT1com);

	p3d_vectNormalize(vecT1T2, vecT1T2);
	p3d_vectNormalize(vecT1com, vecT1com);
	p3d_vectNormalize(vecT2T1, vecT2T1);
	p3d_vectNormalize(vecT2com, vecT2com);

	p3d_vectXprod(vecT1com, vecT1T2, vecRes);
	double d = T1Com * p3d_vectNorm(vecRes);

	double theta1 = p3d_vectDotProd(vecT1T2, vecT1com);
	double theta2 = p3d_vectDotProd(vecT2T1, vecT2com);


	// std::cout << "     DEBUG : " << std::endl
	// 		  << "     Position of com : (" << com[0] << "," << com[1] << ")" << std::endl
	// 		  << "     Position of t1 : (" << t1[0] << "," << t1[1] << ")" << std::endl
	// 		  << "     Position of t2 : (" << t2[0] << "," << t2[1] << ")" << std::endl
	// 		  << "     Coordinate of vector vecT1T2 : (" << vecT1T2[0] << "," << vecT1T2[1] << "," << vecT1T2[2] << ")" << std::endl
	// 		  << "     Coordinate of vector vecT1com : (" << vecT1com[0] << "," << vecT1com[1] << "," << vecT1com[2] << ")" << std::endl
	// 		  << "     Coordinate of vector vecT2T1 : (" << vecT2T1[0] << "," << vecT2T1[1] << "," << vecT2T1[2] << ")" << std::endl
	// 		  << "     Coordinate of vector vecT2com : (" << vecT2com[0] << "," << vecT2com[1] << "," << vecT2com[2] << ")" << std::endl
	// 		  << "     Theta 1 = " << theta1 << ", Theta 2 = " << theta2 << ", d = " << d << std::endl
	// 		  << std::endl;


	if (theta1 < -0.3 || theta2 < -0.3)
	{
		std::cout << "ERROR : projection of COM is outside the center-of-Ankle segment (" << theta1 << "|" << theta2 << ")" << std::endl;
		return false;
	}

	// Length of the foot is 0.24m, width is 0.12m
	// distance of the projected COM to the center-center segment should be less than half the smallest dimension + some margin
	// 0.24 / 2 = 0.12
	// 0.12 / 2 = 0.06
	// 0.12 + 50% margin is 0.06
	if (d > 0.06)
	{
		std::cout << "ERROR : the distance between COM projection and center-of-Ankle segment is above threshold (6 cm) : " << d << std::endl;
		return false;
	}

	return true;
}

#endif /* HPP */
