#ifndef GEOMETRICTOOLS_HPP
#define GEOMETRICTOOLS_HPP

#include "P3d-pkg.h"
#include "API/Device/robot.hpp"

using namespace std;

class ParallelepipedForm;
class GeometricComputations
{
public:
    GeometricComputations();

    p3d_point getRandomPointInTriangle(p3d_point A, p3d_point B, p3d_point C);


    /**
      *a----b
      *|    |
      *c----d   .p
      *b and c should be opposite
      */
    p3d_point getRandomPointInRectangle(p3d_point A, p3d_point B, p3d_point C, p3d_point D);

    vector<p3d_point> getLowerPointsInObject(Robot* obj, double steepness);
    vector<p3d_point> getUpperPointsInObject(Robot* obj, double steepness);
    p3d_point getCenterOfObj(Robot* obje);

    double compute_triangle_area_x_y2(p3d_point a, p3d_point b, p3d_point c);

    /**
      *a----b
      *|    |
      *c----d   .x
      *b and c should be opposite
      */
    bool point_in_rect2(p3d_point a, p3d_point b, p3d_point c, p3d_point d, p3d_point x);
    bool point_in_segment2d(p3d_point a, p3d_point b, p3d_point x);

    double distBetweenPoints(p3d_point a, p3d_point b);

    /**
      * distance between the points in the p3d_matrix4
      */
    double distBetweenMatrix4(p3d_matrix4 a, p3d_matrix4 b);

    double distBetweenPoints2D(p3d_point a, p3d_point b);

    /**
      * test if four points of hexahedron y are in hexahedron x
      */
    bool hexahedron_in_hexahedron(p3d_point x1,p3d_point x2, p3d_point x3, p3d_point x4,
                                  p3d_point x5, p3d_point x6, p3d_point x7, p3d_point x8,
                                  p3d_point y1,p3d_point y2, p3d_point y3, p3d_point y4,
                                  p3d_point y5, p3d_point y6, p3d_point y7, p3d_point y8);

    /**  c----d
      * /|   /|
      *a----b h
      *|/g  |/
      *e----f   .x
      *b and c should be opposite
      */
    bool point_in_hexahedron(p3d_point a, p3d_point b, p3d_point c, p3d_point d,
                             p3d_point e, p3d_point f, p3d_point g, p3d_point h,
                             p3d_point x);

    /**
      *a----b
      *| .h |
      *c----d
      *b and c should be opposite
      */
    double compute_pyramid_area(p3d_point a, p3d_point b, p3d_point c, p3d_point d, p3d_point h);

    vector<p3d_point> getRandomPointsInHexahedron(double nbPoints, p3d_point a, p3d_point b, p3d_point c, p3d_point d, p3d_point e, p3d_point f, p3d_point g, p3d_point h);

    p3d_point getClosestBBPointFromPoint(Robot* e, p3d_point s);

    ParallelepipedForm getBoxFromObjects(vector<Robot*> vr, Robot* gripper);

    double distConfs(tr1::shared_ptr<Configuration> q1, tr1::shared_ptr<Configuration> q2);

    vector<p3d_point> getSampled(vector<p3d_point> vp, double sampleRate);

    double computeCost(vector<p3d_point> vHP, vector<p3d_point> vp, double angleWeight, double distHWeight, double distWeight);

    vector<p3d_point> smoothTraj(vector<p3d_point> origTraj, vector<p3d_point> vHP, Robot *r, int nbIterMax, double gainLim, double trajDistWeight, double distToHWeight, double angleWeight);

};


class GeometricForm
{
public:
    GeometricForm() {}

    virtual bool isPointInForm(p3d_point p) = 0;

    virtual double getDistanceFromPoint(p3d_point p) = 0;
    virtual double getPlannarDistanceFromPoint(p3d_point p) = 0;

    virtual p3d_point getRandomPointInForm() = 0;

    virtual vector<p3d_point> getFormGrid(double sampleRate) = 0;

    string type;
    Robot* object;

    bool isSupport;
    map<string,GeometricForm*> supportAreas;


};

class RectangleForm : public GeometricForm
{
public:
    RectangleForm(p3d_point a, p3d_point b, p3d_point c, p3d_point d, p3d_matrix3 norm);

    bool isPointInForm(p3d_point p);

    double getDistanceFromPoint(p3d_point p);
    double getPlannarDistanceFromPoint(p3d_point p);
    p3d_point getRandomPointInForm();
    vector<p3d_point> getFormGrid(double sampleRate);


    p3d_point _a;
    p3d_point _b;
    p3d_point _c;
    p3d_point _d;
    p3d_matrix3 _norm;

};

class TriangleForm : public GeometricForm
{
public:
    TriangleForm(p3d_point a, p3d_point b, p3d_point c, p3d_matrix3 norm);
    bool isPointInForm(p3d_point p);

    double getDistanceFromPoint(p3d_point p);
    double getPlannarDistanceFromPoint(p3d_point p);

    p3d_point getRandomPointInForm();
    vector<p3d_point> getFormGrid(double sampleRate);

private:
    p3d_point _a;
    p3d_point _b;
    p3d_point _c;
    p3d_matrix3 _norm;
};

class CircleForm : public GeometricForm
{
public:
    CircleForm(p3d_point c, double r, p3d_matrix3 norm);
    bool isPointInForm(p3d_point p);

    double getDistanceFromPoint(p3d_point p);
    double getPlannarDistanceFromPoint(p3d_point p);

    p3d_point getRandomPointInForm();
    vector<p3d_point> getFormGrid(double sampleRate);

private:
    p3d_point _c;
    double _r;
    p3d_matrix3 _norm;
};

class ParallelepipedForm : public GeometricForm
{
public:
    ParallelepipedForm(p3d_point a, p3d_point b, p3d_point c, p3d_point d,
                       p3d_point e, p3d_point f, p3d_point g, p3d_point h);
    bool isPointInForm(p3d_point p);

    double getDistanceFromPoint(p3d_point p);
    double getPlannarDistanceFromPoint(p3d_point p);

    p3d_point getRandomPointInForm();
    vector<p3d_point> getFormGrid(double sampleRate);

private:
    p3d_point _a;
    p3d_point _b;
    p3d_point _c;
    p3d_point _d;
    p3d_point _e;
    p3d_point _f;
    p3d_point _g;
    p3d_point _h;
};


#endif // GEOMETRICTOOLS_HPP
