#include "HRICS_distance2d.h"
#include "utils/Geometry.h"

namespace HRICS {

Distance2D::Distance2D(Robot* rob, std::vector<Robot*> humans,std::vector<Joint*> costJoints):
    _rob(rob), _cost_joints(costJoints), _humans(humans),
    _ignore_threshold(5),
    _minCost(1),
    _maxCost(100),
    _maxToMeanRatio(0.5)
{
    initHumanPos();
}

Distance2D::Distance2D(bool guess_settings){
    if(guess_settings){
        for (unsigned int i=0;i< global_Project->getActiveScene()->getNumberOfRobots() ; i++){
            Robot *r=global_Project->getActiveScene()->getRobot(i);
            if(r->getHriAgent()->is_human){
                _humans.push_back(r);
            }
        }
        initHumanPos();
        setRobot(global_Project->getActiveScene()->getActiveRobot());
        //TODO set joints cost
    }
}

Distance2D::~Distance2D()
{

}


void Distance2D::initHumanPos(){
    clearHumanPos();
    _human_pos.reserve(_humans.size());

    for(unsigned int i=0;i<_humans.size();i++){
        _human_pos.push_back(m3dGeometry::getConfBase2DPos(*_humans[i]->getCurrentPos()));
    }
}

void Distance2D::clearHumanPos(){
    _human_pos.clear();
}

double Distance2D::getCostMax(){
    double cost(0);
    double dist(numeric_limits<double>::infinity());

    for (unsigned int i=0; i<_cost_joints.size() ; i++){
        Joint* jnt=_cost_joints[i];
        Eigen::Vector3d jpos = jnt->getVectorPos();
        Eigen::Vector2d jvect; //2D
        jvect[0]=(jpos.x()); jvect[1] = jpos.y();

        for( unsigned int h=0;h<_human_pos.size();h++){
            dist = std::min(dist, (jvect - _human_pos[h]).norm());
        }


    }

    if(_cost_joints.size()==0){
        cout <<"no joints specified to compute cost" <<endl;
    }

    dist = std::min(dist,_ignore_threshold); //saturate the distance at _ignore_threshold
    cost = _maxCost + (_minCost - _maxCost)/_ignore_threshold * dist; //linear slope from _maxCost to _minCost for distance from 0 to _ignore_threshold
    return cost;
}

double Distance2D::getCostMean(){
    double cost(0);
    double sum(0),mean(0);

    for (unsigned int i=0; i<_cost_joints.size() ; i++){
        Joint* jnt=_cost_joints[i];
        Eigen::Vector3d jpos = jnt->getVectorPos();
        Eigen::Vector2d jvect; //2D
        jvect[0]=(jpos.x()); jvect[1] = jpos.y();
        for( unsigned int h=0;h<_human_pos.size();h++){
            sum = sum + (jvect - _human_pos[h]).norm();
        }
    }

    if(_cost_joints.size()){
        mean = sum / _cost_joints.size();
    }else{
        mean=_ignore_threshold; //leads to the minimal cost
        cout <<"no joints specified to compute cost"<<endl;
    }

    mean = std::min(mean,_ignore_threshold); //saturate the distance at _ignore_threshold
    cost = _maxCost + (_minCost - _maxCost)/_ignore_threshold * mean;
    return cost;
}

double Distance2D::getCost(){
    double cost_mean(0), cost_max(0);
    double mindist(numeric_limits<double>::infinity());
    double sum(0),mean(0);

    for (unsigned int i=0; i<_cost_joints.size() ; i++){
        Joint* jnt=_cost_joints[i];
        Eigen::Vector3d jpos = jnt->getVectorPos();
        Eigen::Vector2d jvect; //2D
        jvect[0]=(jpos.x()); jvect[1] = jpos.y();
        for( unsigned int h=0;h<_human_pos.size();h++){
            sum = sum + (jvect - _human_pos[h]).norm();
            mindist = std::min(mindist, (jvect - _human_pos[h]).norm());
        }
    }

    if(_cost_joints.size()){
        mean = sum / _cost_joints.size();
    }else{
        mean=_ignore_threshold; //leads to the minimal cost
        cout<<"no joints specified to compute cost"<<endl;
    }

    mean = std::min(mean,_ignore_threshold); //saturate the distance at _ignore_threshold
    mindist = std::min(mindist,_ignore_threshold); //saturate the distance at _ignore_threshold
    cost_mean = _maxCost + (_minCost - _maxCost)/_ignore_threshold * mean;
    cost_max = _maxCost + (_minCost - _maxCost)/_ignore_threshold * mindist;
    assert(_maxToMeanRatio <=1.);
    return _maxToMeanRatio * cost_max + (1-_maxToMeanRatio) * cost_mean;
}

void Distance2D::setMaxToMeanRatio(double a)
{
    _maxToMeanRatio = std::min(1.,std::max(0.,a));
    if (a!=_maxToMeanRatio){
//        M3D_ERROR("maxToMeanRatio must be in [0,1]");
        throw std::invalid_argument("maxToMeanRatio must be in [0,1]");
    }
}


} // namespace HRICS

