/*
 * RRT.cpp
 *
 *  Created on: Nov 20, 2012
 *      Author: ddevaurs
 */

#include "API/Roadmap/compco.hpp"
#include "API/Roadmap/graph.hpp"

#include "planner/planEnvironment.hpp"
#include "planner/Diffusion/RRT.hpp"
#include <fstream>
#include <iomanip>
#include <time.h>
using namespace std;
using namespace API;


/**
 * ****************************************************
 *      RRT Expansion
 * ****************************************************
 */

/**
 * RRT Extend.
 * @param fromNode: node from which the expansion is performed
 * @param directionConf: configuration toward which the expansion is going
 * @param directionNode: node toward which the expansion is going (when relevant)
 * @return the number of created nodes
 */
unsigned RRTExpansion::extend(Node * fromNode, confPtr_t directionConf, Node * directionNode)
{

    confPtr_t fromConf,dirConf;

    fromConf=fromNode->getConfiguration();
    dirConf=directionConf;

    // path along which the extension is performed
    LocalPath directionPath(fromConf,dirConf);
    double pathLength = directionPath.getParamMax();

    // construct an 'extension path', by interpolation along the 'direction path'
    double param;
    localPathPtr_t extensionPath;
    param=min(this->step(),pathLength);
    extensionPath.reset(new LocalPath(fromConf,directionPath.configAtParam(param)));

    // if the extension path is valid, add a new node to the graph
    if (extensionPath->isValid())
    {
        this->expansionSucceeded(fromNode);
        bool existingNode = (param==pathLength && directionNode);
        this->addNode(extensionPath, fromNode, existingNode ? directionNode : NULL);
        return (existingNode ? 0 : 1);
    }
    else
    {
        this->expansionFailed(fromNode);
        return 0;
    }
}


/**
 * RRT Connect.
 * @param fromNode: node from which the expansion is performed
 * @param directionConf: configuration toward which the expansion is going
 * @param directionNode: node toward which the expansion is going (when relevant)
 * @return the number of created nodes
 */
unsigned RRTExpansion::connect(Node * fromNode, confPtr_t directionConf, Node * directionNode)
{
    // path along which the connection is performed
    confPtr_t fromConf,dirConf;
    fromConf=fromNode->getConfiguration();
    dirConf=directionConf;
    LocalPath directionPath(fromConf,dirConf);

    // size of the expansion step
    double pathLength=directionPath.getParamMax();
    double stepSize=min(this->step(),pathLength);

    double param=stepSize;
    bool success(true);
    confPtr_t currentConf=fromConf;
    localPathPtr_t connectPath;

    // attempt the connection step by step
    while (success &&  param<=pathLength)
    {

        confPtr_t newConf=directionPath.configAtParam(param);
        connectPath.reset(new LocalPath(currentConf,newConf));

        // check whether the current path is valid
        success=connectPath.get()->isValid();

        // if it is go to the next portion
        if (success)
        {
            currentConf=newConf;
            if (param == pathLength)
                param = pathLength + 1.;
            else if (pathLength - param > stepSize + 1e-3)
                param += stepSize;
            else
                param = pathLength;
        }
    }

    // if "reached" or "advanced", add a node to the graph
    if (success || param!=stepSize)
    {
        connectPath.reset(new LocalPath(fromConf,currentConf));
        this->expansionSucceeded(fromNode);
        bool existingNode = ((connectPath.get()->getEnd() == directionPath.getEnd()) && directionNode);
        this->addNode(connectPath, fromNode, existingNode ? directionNode : NULL);
        return (existingNode ? 0 : 1);
    }
    else
    {
        this->expansionFailed(fromNode);
        return 0;
    }
}


/**
 * Try to add cycles to the graph.
 * @return TRUE if at least one cycle has been added, and FALSE otherwise
 */
bool RRTExpansion::addCycles()
{
	unsigned nbCycles(0);
	vector<Node *> nodes = m_Graph->getNodes();
    Node * lastNode = m_Graph->getLastNode();
	confPtr_t lastConf = lastNode->getConfiguration();
	double distanceThreshold = ENV.getDouble(Env::minimalFinalExpansionGap) * this->step();

	// compute the distance (in the graph) between the last node and every other node
	vector<double> graphDistance = m_Graph->computeDijkstraShortestPaths(lastNode);

	for (unsigned i(0); i < nodes.size() - 1; ++i)
	{
		// compute the length of the path (in C-space) between the last node and each candidate node
		double spaceDistance = lastConf->distance(nodes[i]->getConfiguration().get());

		// do not consider nodes that are too far away from the last node
		if (spaceDistance <= distanceThreshold)
		{
			// for a cycle to be considered useful, the distance of the associated path (in C-space)
			// has to be smaller than the distance of the alternative path in the graph
			if (spaceDistance < graphDistance[i])
			{
				localPathPtr_t path(new LocalPath(lastConf, nodes[i]->getConfiguration()));
				if (path->isValid())
				{
					// if the path is valid, add a useful cycle
					m_Graph->addEdges(lastNode, nodes[i], path);
					++nbCycles;
				}
			}
		}
	}

	// return TRUE if at least one cycle has been added, and FALSE otherwise
	return (nbCycles > 0);
}

//Holding space for addCycle strategies in TRRT
bool RRTExpansion::addCycles2()
{
    return this->addCycles();
}
bool RRTExpansion::addCycles3()
{
    return this->addCycles();
}

bool RRTExpansion::addCycles1()
{//Instead of trying to connect only the last added node at its closer neighbors
    //we try to connect all the nodes to their k nodes
    //k is function of the successRate of adding cycle in the tree
    double successRate = 0;
    vector<Node *> nodes = m_Graph->getNodes();

    static int k = 3;
    static double successRatePrev = 0;

    for (unsigned i(0); i < nodes.size() - k; ++i)
    {
        Node * current = nodes.at(i);

        for(uint j = 1 ; j < k;j++)
        {
            Node * newLink = nodes.at(i+j);
            if(!m_Graph->isEdgeInGraph(current, newLink))
            {
                localPathPtr_t path(new LocalPath(current->getConfiguration(), newLink->getConfiguration()));
                if (path->isValid())
                {
                    // if the path is valid, add a useful cycle
                    m_Graph->addEdges(current, newLink, path);
                    successRate += 2/(current->getNodeStruct()->dist_Nnew + newLink->getNodeStruct()->dist_Nnew);
                }
            }

        }

    }

    m_Graph->sortNodes();

    if(successRate == 0 && successRatePrev == 0)
    {
        k += 1;
    }
    else if(successRate > 0 && successRatePrev > 0)
    {
        k -= 1;
    }

    k = (k <= 2)?3:k;

    successRatePrev = successRate;

    // return TRUE if at least one cycle has been added, and FALSE otherwise
    return (successRate > 0);
}


/**
 * Add a node to the graph, after an expansion that has produced the given path, starting from fromNode.
 * If the expansion has successfully reached the already existing node directionNode,
 * directionNode is simply linked to fromNode.
 * Otherwise, a new node containing the end configuration of the path is created and linked to fromNode.
 * @return the added node
 */
Node * RRTExpansion::addNode(localPathPtr_t path, Node * fromNode, Node * directionNode)
{
    // if the expansion was performed toward an already existing node, both components are merged
    if (directionNode)
    {
        m_Graph->addEdges(fromNode,directionNode, path);
        return directionNode;
    }
    // otherwise, a new node is created in the graph and linked to fromNode
    else
    {
        Node * newNode = new Node(m_Graph, path->getEnd());
        m_Graph->addNode(newNode);
        m_Graph->addEdges(fromNode, newNode, path);
        return newNode;
    }
}

/**
 * ****************************************************
 *             RRT
 * ****************************************************
 */


/**
 * Main function of RRT.
 * @return the number of nodes added to the graph
 */
unsigned RRT::run()
{
    ofstream myfile;
    if(PlanEnv->getBool(PlanParam::anytime))
    {//in order to easily evaluate the performance of the different anytime startegies
    //we generate a json file with relevant info
        myfile.open ("./addcycle.json",ios::app);
        myfile<<"{\"values\":["<<endl;
    }
    clock_t t;
    int nAddCycleFail = 0;
    int nAddCycle = 0;
    int nAddCycle1 = 0;
    
    int switchV = 10; //(0,50) The number of failures from a strategy we tolerate before switching to another one if available
    
    int nCycle = 0;

    if(ENV.getBool(Env::armOnly)){
            lm_quadspline_str *param=lm_get_quadspline_lm_param(_Robot->getRobotStruct());
            bool precond=(param!=NULL);
            if(precond){
                configPt qi=_q_start->getConfigStruct();
                configPt qf=_q_goal->getConfigStruct();
                precond=precond && (fabs(qi[param->x_dof]-qf[param->x_dof])<EPS10);
                precond=precond && (fabs(qi[param->y_dof]-qf[param->y_dof])<EPS10);
                precond=precond && (fabs(qi[param->z_dof]-qf[param->z_dof])<EPS10);
                precond=precond && (fabs(qi[param->rx_dof])<EPS10);
                precond=precond && (fabs(qi[param->ry_dof])<EPS10);
                precond=precond && (fabs(qi[param->rz_dof]-qf[param->rz_dof])<EPS10);
                precond=precond && (fabs(qi[param->vx_dof])<EPS10);
                precond=precond && (fabs(qi[param->vy_dof])<EPS10);
                precond=precond && (fabs(qi[param->vz_dof])<EPS10);
                precond=precond && (fabs(qi[param->ax_dof])<EPS10);
                precond=precond && (fabs(qi[param->ay_dof])<EPS10);
                precond=precond && (fabs(qi[param->az_dof])<EPS10);
                precond=precond && (fabs(qf[param->vx_dof])<EPS10);
                precond=precond && (fabs(qf[param->vy_dof])<EPS10);
                precond=precond && (fabs(qf[param->vz_dof])<EPS10);
                precond=precond && (fabs(qf[param->ax_dof])<EPS10);
                precond=precond && (fabs(qf[param->ay_dof])<EPS10);
                precond=precond && (fabs(qf[param->az_dof])<EPS10);
            }
            if(!precond){
                printf("RRT::run: pre-conditions are not satisfied for armOnly planning.\n");
                return 0;
            }
            else{
                printf("RRT::run: armOnly precond OK\n");
            }
            _Robot->getRobotStruct()->lpl_type=LINEAR;
            ENV.setBool(Env::isHybrid,true);
    }

	this->chronoStart();

	// check whether the pre-conditions are satisfied
	if (!this->preConditions()) {
        if (PlanEnv->getBool(PlanParam::showGTPDebug))
            cout << "Failure: pre-conditions are not satisfied." << endl;
		return 0;
	}

	// check whether this is a trivial planning scenario
	localPathPtr_t directPath(new LocalPath(_q_start, _q_goal));
	if (ENV.getBool(Env::expandToGoal) && expansion->isValidConnection(directPath))
	{
		_Graph->addEdges(_Start, _Goal, directPath);
        if (PlanEnv->getBool(PlanParam::showGTPDebug))
            cout << "Trivial: Direct connection between Start and Goal." << endl;
		return 0;
	}

	unsigned nbCreatedNodes(0); // number of nodes created by one expansion step
	unsigned totalCreatedNodes(0); // total number of nodes created so far
	ConnectedComponent * fromComp; // component that will be expanded during the current step
	ConnectedComponent * toComp; // component toward which to expand (when relevant)

	fromComp = _Start->getConnectedComponent();
	toComp = _Goal->getConnectedComponent();

	// variable used to keep track of the cost of the best trajectory found so far, in the anytime variant
	double bestTrajCost = numeric_limits<double>::max();
	bool firstTraj(true);

	// RRT loop: start by checking the stopping conditions
	while (!this->checkStopConditions())
	{
		// in the case of a balanced Bi-directional RRT, do not expand when components are unbalanced
		if (!ENV.getBool(Env::biDir) || !ENV.getBool(Env::expandBalanced) ||
				(fromComp->getNumberOfNodes() < toComp->getNumberOfNodes() + 5))
		{
			++m_nbExpansion;

			// perform a single expansion step of RRT
			nbCreatedNodes = this->expandOneStep(fromComp);

			// success, if some nodes have been created
			if (nbCreatedNodes > 0)
			{
				totalCreatedNodes += nbCreatedNodes;
				m_nbConscutiveFailures = 0;

				Node * lastNode = _Graph->getLastNode();
                
                if (ENV.getBool(Env::addCycles1) && this->trajFound())
                    {
                //_Graph->updateSortedNodes(lastNode);
                _Graph->updateSepareNodes(lastNode);
                    }
				confPtr_t lastConf = lastNode->getConfiguration();

				// attempt a connection toward the goal / the other component (when relevant)
				if (ENV.getBool(Env::expandToGoal) && !this->trajFound())
				{
					// Bi-directional RRT
					if (ENV.getBool(Env::biDir))
                    {
                        // look for the node whose configuration is the closest to the last configuration
                        // within the 'toComp' connected component
                        Node * closestNode=toComp->nearestNeighbor(lastConf);

                        // create local path
                        localPathPtr_t path;
                        path.reset(new LocalPath(lastConf,closestNode->getConfiguration()));

                        // if the path between both components is valid, connect them
                        if (expansion->isValidConnection(path))
                        {
                            _Graph->addEdges(lastNode, closestNode, path);
                            toComp = fromComp = _Start->getConnectedComponent();
                        }
					}

					// mono-directional RRT (with goal)
					else
					{
						// if the path between the last node and the goal is valid, connect them
						localPathPtr_t path(new LocalPath(lastConf, _q_goal));
						if (expansion->isValidConnection(path))
							_Graph->addEdges(lastNode, _Goal, path);
					}
				}

				// anytime variant
				bool drawNewTraj = false;
				if (PlanEnv->getBool(PlanParam::anytime) && this->trajFound())
				{
					nCycle++;
					bool lastNodeUsed = false;
					lastNode->useful = -10000;//a node is useful when it is used in a trajectory
                                              //the info allow graph reduction.
					bool extractNewTraj(true);

                    if(firstTraj)
                    {
                        t = clock();
                    }
					// try to add cycles (when relevant)                    
                    
					if ((ENV.getBool(Env::addCycles) && !ENV.getBool(Env::addCycles1)) || (ENV.getBool(Env::addCycles) && ENV.getBool(Env::addCycles1) && nAddCycleFail >= switchV)) 
                    {
                        extractNewTraj = (firstTraj || expansion->addCycles());
                        nAddCycle++;
                    }
                    else if ((!ENV.getBool(Env::addCycles) && ENV.getBool(Env::addCycles1)) || (ENV.getBool(Env::addCycles) && ENV.getBool(Env::addCycles1) && nAddCycleFail < switchV))
                    {
                        extractNewTraj = (firstTraj || expansion->addCycles1());
                        nAddCycle1++;
                    }
					// extract the new trajectory from the graph 
					if (extractNewTraj)
					{
						Trajectory * trajectory = this->extractTrajectory();
						if (trajectory)
						{
							double trajCost = trajectory->cost();
							if (trajCost < bestTrajCost)
							{
								// keep track of the best trajectory found so far
								bestTrajCost = trajCost;
								trajectory->replaceP3dTraj();
								drawNewTraj = !ENV.getBool(Env::drawDisabled);
                                int duration = clock() - t;
                                cout << switchV << " best trajectory cost = " << bestTrajCost << endl;
                                myfile<<"{\"duration\" : "<<((float)duration)/CLOCKS_PER_SEC<<",";
                                myfile<<"\"nodes\" : "<<_Graph->getNodes().size()<<",";
                                myfile<<"\"edges\" : "<<_Graph->getEdges().size()<<",";
                                myfile<<"\"nAddCycle\" : "<<nAddCycle<<",";
                                myfile<<"\"nAddCycle1\" : "<<nAddCycle1<<",";
                                myfile<<"\"cost\" : "<<std::setprecision(16)<<bestTrajCost<<"},"<<endl;

                                this->trajCosts.push_back(bestTrajCost);
                                trajToDraw.push_back(*trajectory);
                                t = clock();

                                if(ENV.getBool(Env::addCycles1)){
                                    _Graph->separateNodes();
                                    nAddCycleFail = (nAddCycleFail > switchV)?0:nAddCycleFail-1;
									if(!firstTraj)
									{
                                        //switchV can be computed in various way : here are 3 that are quite good (empirical results)
										//switchV = (int) (22*log(1+100*(1- bestTrajCost/this->trajCosts.at(this->trajCosts.size() - 2))));
										//switchV = (int) (9+ exp((11+ 100*(1 - bestTrajCost/this->trajCosts.at(this->trajCosts.size() - 2)))/24));
										switchV = (int) (10 + 100*(1 - bestTrajCost/this->trajCosts.at(this->trajCosts.size() - 2)));  
									}
                                }
                                
                                nAddCycle = 0;
                                nAddCycle1 = 0;

                                
                                vector<Node *> trajNode = _Graph->extractAStarShortestNodePaths(_Start, _Goal);
                                //updating the usefulness of the nodes
                                for(uint i = 0; i < trajNode.size(); i++)
                                {
									if( trajNode[i]->useful <= -10000 )
                                    {	
										trajNode[i]->useful += nCycle;
									}
                                    if(trajNode[i] == lastNode){
                                        lastNodeUsed = true;
                                    }
                                }
							}

                            firstTraj = false;
                            
						}
					}
                    else
                    {
                        nAddCycleFail++;
                    }
                    
                    
                    if(!lastNodeUsed)
                    {
						lastNode->useful -= nCycle;
					}
				}

				// redraw the scene (when relevant)
				if (!ENV.getBool(Env::drawDisabled) && (ENV.getBool(Env::drawExploration) || drawNewTraj))
					_draw_func();
			}
			// failure, otherwise
			else {
				++m_nbFailedExpansion;
				++m_nbConscutiveFailures;
			}
		}
		this->chronoUpdate();

		// in the case of the Bi-directional RRT, swap the roles of the connected components
        if (ENV.getBool(Env::biDir))
            swap(fromComp, toComp);
	}

	this->chronoStop();
    
    if(PlanEnv->getBool(PlanParam::anytime))
    {//exporting data
        vector<Node *> nodes = _Graph->getNodes();

        int nonUsefulNodes = 0;
        int usefulNodes = 0;
        int cycleNeeded = 0;
        for(uint i = 0; i<nodes.size(); i++)
        {
            if(nodes[i]->useful < -10000)
            {
                nonUsefulNodes++;
            }
            else if(nodes[i]->useful >= -10000 && nodes[i]->useful < 0)
            {
                usefulNodes++;
                cycleNeeded+= nodes[i]->useful + 10000;
            }
        }
        
        if(usefulNodes > 0)
    		cycleNeeded /= usefulNodes;
        
        int duration = clock() - t;
                                    myfile<<"{\"duration\" : "<<((float)duration)/CLOCKS_PER_SEC<<",";
                                    myfile<<"\"nodes\" : "<<_Graph->getNodes().size()<<",";
                                    myfile<<"\"edges\" : "<<_Graph->getEdges().size()<<",";
                                    myfile<<"\"nAddCycle\" : "<<nAddCycle<<",";
                                    myfile<<"\"nAddCycle1\" : "<<nAddCycle1<<",";
                                    myfile<<"\"cost\" : "<<std::setprecision(16)<<bestTrajCost<<"}"<<endl;
            myfile<<"], \"stats\" : {\"NonUseful\" : "<<nonUsefulNodes<<", \"Useful\" : "<<usefulNodes<<", \"cycleNeeded\":"<<cycleNeeded<<",\"ncycle\":"<<nCycle<<",\"TotalNodes\" : "<<nodes.size()<<"}},"<<endl;
            myfile.close();
    }
	return totalCreatedNodes;
}



/**
 * Check whether RRT has found a trajectory in the graph it is building.
 * @return TRUE if the criteria required for a trajectory to be found are met, and FALSE otherwise
 */
bool RRT::trajFound()
{
	// success if the Start and Goal are in the same component
	return (_Start->inSameComponent(_Goal));
}


/**
 * Perform a single expansion step of RRT, growing the given connected component.
 * @return the number of created nodes
 */
unsigned RRT::expandOneStep(ConnectedComponent * fromComp)
{
	// sample the expansion direction (i.e. the configuration toward which the expansion will be performed)
	// taking into account the goal bias, if relevant
	Node * directionNode(NULL);
    confPtr_t directionConf;

	bool authorizedBias = !PlanEnv->getBool(PlanParam::anytime) || !this->trajFound();
	if (ENV.getBool(Env::isGoalBiased) && authorizedBias && this->goalBias())
	{
		// bias toward the goal configuration
		directionNode = _Goal;
		directionConf = _q_goal;
	}
	else
		// or random sampling in the entire C-Space
		directionConf = _Robot->shoot();

	// select, from the component to be grown, the node from which the expansion will be attempted
	Node * expandNode = ENV.getBool(Env::randomConnectionToGoal) ? fromComp->randomNode()
			: fromComp->nearestNeighbor(directionConf);

    // In case of a plannification only for the arm of the quadrotor
    if(ENV.getBool(Env::armOnly)){
        lm_quadspline_str *param=lm_get_quadspline_lm_param(_Robot->getRobotStruct());
        if(!param){
            printf("error: RRT::expandOneStep: no quadspline structure initialized for robot %s\n",_Robot->getName().c_str());
            return 0;
        }
        configPt qi=expandNode->getConfiguration()->getConfigStruct();
        configPt qf=directionConf->getConfigStruct();
        qf[param->x_dof]=qi[param->x_dof];
        qf[param->y_dof]=qi[param->y_dof];
        qf[param->z_dof]=qi[param->z_dof];
        qf[param->rx_dof]=0;
        qf[param->ry_dof]=0;
        qf[param->rz_dof]=qi[param->rz_dof];
        qf[param->vx_dof]=0;
        qf[param->vy_dof]=0;
        qf[param->vz_dof]=0;
        qf[param->ax_dof]=0;
        qf[param->ay_dof]=0;
        qf[param->az_dof]=0;
    }

	// directed expansion
	switch (ENV.getExpansionMethod())
	{
	case Env::Extend:
		return expansion->extend(expandNode, directionConf, directionNode);
	case Env::Connect:
		return expansion->connect(expandNode, directionConf, directionNode);
	default:
		cerr << "*** ERROR: RRT::expandOneStep --> method not implemented !!! ***" << endl;
		return 0;
	}
}


/**
 * Check the stopping conditions of RRT.
 * @return TRUE if one of the conditions is satisfied, and FALSE otherwise
 */
bool RRT::checkStopConditions()
{
	if (TreePlanner::checkStopConditions())
    {
		return true;
    }

	if (ENV.getBool(Env::expandToGoal) && !PlanEnv->getBool(PlanParam::anytime))
		return (this->trajFound());

	return false;
}


/**
 * Extract a trajectory from Start to Goal.
 */
Trajectory * RRT::extractTrajectory()
{
	if (ENV.getBool(Env::biDir) || PlanEnv->getBool(PlanParam::anytime))
		return _Graph->extractAStarShortestPath(_Start, _Goal);
	else
		return _Graph->extractTreeShortestPath(_Start, _Goal);
}

