/*
 *  Replanner.cpp
 *  libmove3D-planners 
 *
 *  Created by Benjamin Vadant on 01/02/13.
 *  Based on Jim Mainprice replanningAlgorithms.hpp/cpp
 *  Copyright 2013 LAAS/CNRS. All rights reserved.
 *
 */

#include "Replanner.hpp"
#include "ReplanningSimulator.hpp"

#include "API/project.hpp"

#include "planner/planEnvironment.hpp"
#include "planner/plannerFunctions.hpp"
#include "planner/TrajectoryOptim/Stomp/stompOptimizer.hpp"
#include "planner/TrajectoryOptim/trajectoryOptim.hpp"
#include "planner/TrajectoryOptim/Classic/costOptimization.hpp"
#include "planner/cost_space.hpp"
#include "planner/Diffusion/Variants/RRT-Star.hpp"

#include "HRICS_costspace.hpp"

#include "LightPlanner-pkg.h"
#include "Graphic-pkg.h"
#include "Collision-pkg.h"

#include "move3d-headless.h"

#include <boost/thread/thread.hpp>
#include <sys/time.h>

using namespace std;
using namespace tr1;

//! Deactivate all kinematic constraints
void p3d_deactivate_all_cntrts( Robot* r )
{
    p3d_rob* rob = r->getRobotStruct();
    
    int nb_cntrts = rob->cntrt_manager->ncntrts;
    p3d_cntrt** cntrts = rob->cntrt_manager->cntrts;
    
    for(int i=0; i<nb_cntrts; i++) 
        p3d_desactivateCntrt( rob , cntrts[i] );
}

//! Initialize replanner by setting the robot, the multi-localpaths available
//! and calling the init function
Replanner::Replanner(Robot* r) : robot(r), human(NULL), useMultiLocalPath(false)
{
    cout << "New base replanner" << endl;
    
    if( !this->initMultiLocalPath() )
    {
        cout << "Error initializing multi-localpahts in " << __FILE__ << " at " << __func__ << endl;
    }
}

Replanner::~Replanner() 
{

}

void Replanner::setSwitchData( confPtr_t qSwitch, int idSwitch, double positionSwitch, double timeLimitRePlanning, double localPathAverageLength, double initialStep )
{
    this->qSwitch                = qSwitch;
    this->idSwitch               = idSwitch;
    this->positionSwitch         = positionSwitch;
    this->timeLimitRePlanning    = timeLimitRePlanning;
    this->localPathAverageLength = localPathAverageLength;
    this->initialStep            = initialStep;
}

//! Initialize the robot multilocal path
//! disable the display of the main robot
bool Replanner::initMultiLocalPath()
{
    
    if (robot == NULL)
        return false;
    
    for (int i=0; i<robot->getRobotStruct()->mlp->nblpGp; i++) 
    {
        if (!strcmp(robot->getRobotStruct()->mlp->mlpJoints[i]->gpName, "base")) {
            baseMultiLocalPath = i;
            useMultiLocalPath = true;
        } else if (!strcmp(robot->getRobotStruct()->mlp->mlpJoints[i]->gpName, "baseSm")) {
            baseSoftMotionMultiLocalPath = i;
            useMultiLocalPath = true;
        } else if (!strcmp(robot->getRobotStruct()->mlp->mlpJoints[i]->gpName, "head")) {
            headMultiLocalPath = i;
            useMultiLocalPath = true;
        } else if (!strcmp(robot->getRobotStruct()->mlp->mlpJoints[i]->gpName, "upBody")) {
            upBodyMultiLocalPath = i;
            useMultiLocalPath = true;
        } else if (!strcmp(robot->getRobotStruct()->mlp->mlpJoints[i]->gpName, "upBodySm")) {
            upBodySoftMotionMultiLocalPath = i;
            useMultiLocalPath = true;
        }
    }
    
    cout << "Replanner::robot : " << robot->getName() << endl;
    cout << "Replanner::useMultiLocalPath : " << useMultiLocalPath << endl;
    return useMultiLocalPath;
}

//! Concatenate the new portion of the trajectory which is going to be used 
//! The concatanation happens at swith id which has to be computed before
std::pair<bool,API::Trajectory>  Replanner::concatenateToCurrentTrajectory(const API::Trajectory& newPortion)
{
    pair<bool,API::Trajectory> pair;
    pair.first = false;
    pair.second = currentTrajectory;
    
    if ( robot==NULL ) {
        cout <<  "concatenateToCurrentTrajectory: robot is NULL.\n";
        return pair;
    }
    
    API::Trajectory concatenatedTrajectory = currentTrajectory.extractSubTrajectory( 0, positionSwitch, false );
    
    cout << "concatenatedTrajectory.getRangeMax() before concatenation : " << concatenatedTrajectory.getRangeMax() << endl;
    if( concatenatedTrajectory.concat( newPortion ) ) {
        pair.first = true;
        pair.second = concatenatedTrajectory;
    }
    cout << "concatenatedTrajectory.getRangeMax() after concatenation :  " << concatenatedTrajectory.getRangeMax() << endl;
    return pair;
}

/*//! Concatenate the multiple portions of trajectories which are going to be used
//! The concatenation happens at idSwitch which has to be computed before
std::pair<bool,API::Trajectory>  Replanner::concatenateToCurrentTrajectory(const vector<API::Trajectory> trajectories)
{
    pair<bool,API::Trajectory> pair;
    pair.first = false;
    pair.second = currentTrajectory;
    
    if ( robot==NULL ) {
        cout <<  "concatenateToCurrentTrajectory: robot is NULL.\n";
        return pair;
    }
    if ( trajectories.empty()) {
        cout <<  "concatenateToCurrentTrajectory: the trajectory vector is empty.\n" ;
        return pair;
    }
    
    API::Trajectory concatenatedTrajectory = currentTrajectory.extractSubTrajectory( 0, positionSwitch, false );
    
    for (int i=0; i<int(trajectories.size()); i++) {
        if(!trajectories[i].isEmpty()) {
            cout << "concatenatedTrajectory.getRangeMax() before concatenation : " << concatenatedTrajectory.getRangeMax() << endl;
            if( concatenatedTrajectory.concat( trajectories[i] ) ) {
                pair.first = true;
                pair.second = concatenatedTrajectory;
            }
            cout << "concatenatedTrajectory.getRangeMax() after concatenation :  " << concatenatedTrajectory.getRangeMax() << endl;
        }
    }
    return pair;
}*/

//DEPRECATED
//! Concatenate the multiple portions of trajectories which are going to be used 
//! The concatenation happens at idSwitch which has to be computed before
p3d_traj* Replanner::concatenateToCurrentTrajectory(const vector<p3d_traj*>& trajectories)
{
    if ( robot==NULL ) {
        cout <<  "concatenateToCurrentTrajectory: robot is NULL.\n";
        return NULL;
    }
    if ( trajectories.empty() || !trajectories[0] ) {
        cout <<  "concatenateToCurrentTrajectory: the trajectory vector is empty.\n" ;
        return NULL;
    }
    
    // if the currentTrajectory is not empty
    // concat the allready executed part begore idSwitch to the new trajectory
    int init_id;
    p3d_traj* concatTraj;
    if ( !currentTrajectory.isEmpty() ) {
        concatTraj = currentTrajectory.extractSubTrajectoryOfLocalPaths(0,idSwitch).replaceP3dTraj(NULL);
        init_id = 0; 
    }
    else {
        concatTraj = trajectories[0];
        init_id = 1; 
    }
    
    for (int i=init_id; i<int(trajectories.size()); i++) {
        if(trajectories[i]) {
            if (p3d_concat_traj( concatTraj, trajectories[i]) == TRUE) {
                cout << "Concat traj fails" << endl;  
                return NULL;
            }
        }
    }
    
    return concatTraj;
}
