#include "MultiHandOverUtils.hpp"
#include "planner/planner.hpp"
#include "HRI_costspace/HRICS_MultiHandOver.h"

#include "planner/TrajectoryOptim/plannarTrajectorySmoothing.hpp"
#include "API/project.hpp"
#include "plannerFunctions.hpp"
#include "Planner-pkg.h"
#include "LightPlanner-pkg.h"
#include "multiHandOver/graphAlgorithms.h"
#include "MultiHandOverAlgos.h"
#include <sys/time.h>
#include <boost/property_tree/xml_parser.hpp>
#include <boost/property_tree/ptree.hpp>
#include <boost/foreach.hpp>
#include <boost/algorithm/string/predicate.hpp>

#include "Collision-pkg.h"


using namespace std;
namespace HRICS {
namespace MHO {

ConfigFileParser::ConfigFileParser(MultiHandOver *mho):
    _mho(mho)
{}

bool ConfigFileParser::configure_from_xml(istream &is){
    //populate tree structure
    using namespace boost::property_tree;
    bool ok(0);
    ptree pt;
    try{
        xml_parser::read_xml(is,pt);
    }catch(xml_parser_error e){
        cerr << "MHO::ConfigFileParser (XML) error parsing xml file: " << e.what()<<endl;
        return false;
    }catch(std::exception e){
        cerr << "MHO::ConfigFileParser (XML) error : "<< e.what() <<endl;
        return false;
    }

    //traverse pt
    BOOST_FOREACH (ptree::value_type const& v,pt.get_child("MHOSettings")){
        ok=true;
        if(v.first == "agent"){
            double speed,use;
            string name;
            speed=v.second.get<double>("<xmlattr>.speed",1);
            try{
                use=v.second.get<double>("<xmlattr>.use_cost");
            }catch (ptree_bad_data d){
                use=-1;//inf
            }catch (ptree_bad_path p){
                use=1;//default
            }

            name=v.second.get<string>("<xmlattr>.name");
            if(use<0){
                use=numeric_limits<double>::infinity();
            }
            for (unsigned int a=0;a<_mho->getAgents().size();++a){
                if(_mho->getAgentStruct(a)->robot->getName() == name){
                    _mho->setAgentSpeed(a,speed);
                    _mho->setAgentUseCost(a,use);
                    break;
                }
            }
        }else if(v.first == "stop_conditions"){
            int it(-1),oh(-1),t_ms(-1);
            double t_s;
            it=v.second.get<int>("<xmlattr>.iterations",-1);
            oh=v.second.get<int>("<xmlattr>.overheat",-1);
            t_s=v.second.get<double>("<xmlattr>.time",-1.);
            if(t_s>0){
                t_ms=t_s*1000;
            }
            _mho->setSearchSolutionStopCondition(it,oh,t_ms);
        }else if(v.first == "from_to"){
            string name_f,name_t,name_o;
            name_f=v.second.get<string>("<xmlattr>.from");
            name_t=v.second.get<string>("<xmlattr>.to");
            name_o=v.second.get<string>("<xmlattr>.object");
            int randomize = v.second.get<int>("<xmlattr>.randomize",-1);
            if(randomize >=0){
                _mho->setRandomizeFirstTarget(randomize > 0);
            }
            bool f_set(0),t_set(0);
            for (unsigned int a=0;a<_mho->getAgents().size();++a){
                if(!f_set && _mho->getAgentStruct(a)->robot->getName() == name_f){
                    _mho->setFirstAgent(_mho->getAgents()[a]);
                    f_set=true;
                }
                if(!t_set && _mho->getAgentStruct(a)->robot->getName() == name_t){
                    _mho->setTargetAgent(_mho->getAgents()[a]);
                    t_set=true;
                }
                if(t_set && f_set){break;}
            }
            for(unsigned int o=0;o<_mho->getManipulableObjects().size();o++){
                if(_mho->getManipulableObjects()[o]->getName()==name_o){
                    _mho->setObject(o);
                    break;
                }
            }

        }else if(v.first == "cost_factors"){
            double d,h,t,u;
            d=v.second.get<double>("<xmlattr>.distance",1);
            h=v.second.get<double>("<xmlattr>.hri",1);
            t=v.second.get<double>("<xmlattr>.time",1);
            u=v.second.get<double>("<xmlattr>.agent",1);
            _mho->setCostsFactors(d,t,u,h);
        }else if(v.first == "algo_opts"){
            string algo_name, heuristic_name;
            double epsilon(0);
            algo_name=v.second.get<string>("search_algo.<xmlattr>.name",string(""));
            heuristic_name=v.second.get<string>("search_algo.<xmlattr>.heuristic",string(""));
            epsilon=v.second.get<double>("lwa_star.<xmlattr>.epsilon",0.);
            int use_post_process = v.second.get<int>("post_process",-1);
            if(boost::iequals(algo_name,"lwa")){
                _mho->setSearchAlgo(MultiHandOver::LWA_SEARCH);
            }else if(boost::iequals(algo_name,"wa")){
                _mho->setSearchAlgo(MultiHandOver::WA_SEARCH);
            }else if(boost::iequals(algo_name,"random")){
                _mho->setSearchAlgo(MultiHandOver::RANDOM_SEARCH);
            }
            if (boost::iequals(heuristic_name,"dist_best")){
                _mho->setAstarHeuristic(MultiHandOver::DIST_BEST_AGENT);
            }else if (boost::iequals(heuristic_name,"best_ho_series")){
                _mho->setAstarHeuristic(MultiHandOver::BEST_HANDOVER_SERIES);
            }
            if(epsilon){
                _mho->setEpsilonForLWAstar(epsilon);
            }
            if(use_post_process >= 0){
                _mho->usePostProcess(use_post_process);
            }
            //<frontier_search rrt="1" dummy_smooth="1"/> <!-- use RRT for frontier checking or
            int rrt,dummy_smooth;

            rrt=v.second.get<int>("frontier_search.<xmlattr>.rrt",-1);
            dummy_smooth=v.second.get<int>("frontier_search.<xmlattr>.dummy_smooth",-1);
            if(rrt>=0){
                _mho->setUseRrtInFrontierCheck(rrt);
            }
            if(dummy_smooth>=0){
                _mho->setUseDummySmoothInFrontierCheck(dummy_smooth);
            }

            //grid options
            double cell_size;
            cell_size=v.second.get<double>("grid.<xmlattr>.cellsize",-1);
            if(cell_size>0){
                _mho->setGridCellSize(cell_size);
            }
        }else if (v.first == "handover_testers"){
            vector<string> defaults;
            try{
                BOOST_FOREACH(ptree::value_type &v2,
                              pt.get_child("MHOSettings.handover_testers")){
                    if(v2.first=="default"){
                        defaults.push_back(v2.second.data());
                    }else if(v2.first=="check_before"){
                        _mho->setCheckFrontierBefore(v2.second.get<int>("",1)>0);
                    }else if (v2.first == "max_level"){
                        if(boost::iequals(v2.second.data(), "handover")){
                            _mho->setFrontierMaxCheckStatus(FrontierCells::FCS_HANDOVER);
                        }else if (boost::iequals(v2.second.data(),"trajectory")){
                            _mho->setFrontierMaxCheckStatus(FrontierCells::FCS_TRAJECTORIES);
                        }else if (boost::iequals(v2.second.data(),"grab_ik")){
                            _mho->setFrontierMaxCheckStatus(FrontierCells::FCS_GRAB_CONFS);
                        }else{
                            cout<< "MHO WARNING while parsing xml file : handover_testers/max_level '"<<v2.second.data()<<"' not managed"<<endl;
                        }
                    }
                }
            }catch (std::exception e){
                cerr <<e.what();
            }

            _mho->setDefaultHandOverTesters(defaults);
        }else if(v.first == "handover_times"){
            cout<<"getting ho times"<<endl;
            _mho->setTimesHO(v.second.get<double>("<xmlattr>.RR",5),v.second.get<double>("<xmlattr>.HH",2),v.second.get<double>("<xmlattr>.RH",3.5));
        }else{
            cout << "WARNING : xml contains unknown field : \"" <<v.first<<"\""<<endl;
        }
    }
    return ok;
}

bool is_not_digit(char c){return !std::isdigit(c);}

class StdOutMuter{
public:
    static inline void mute(){
        StdOutMuter::cout_sbuf = std::cout.rdbuf(); // save original sbuf
        StdOutMuter::cerr_sbuf = std::cerr.rdbuf(); // save original sbuf
        std::ofstream   fout("/dev/null");
        std::cout.rdbuf(fout.rdbuf()); // redirect 'cout' to a 'fout'
        std::cerr.rdbuf(fout.rdbuf()); // redirect 'cout' to a 'fout'
    }
    static inline void unmute(){
        std::cout.rdbuf(StdOutMuter::cout_sbuf); // restore the original stream buffer
        std::cerr.rdbuf(StdOutMuter::cerr_sbuf); // restore the original stream buffer
    }
    static std::streambuf* cout_sbuf;
    static std::streambuf* cerr_sbuf;
};
std::streambuf* StdOutMuter::cout_sbuf=std::cout.rdbuf();
std::streambuf* StdOutMuter::cerr_sbuf=std::cout.rdbuf();

MultiHandOverAgent::MultiHandOverAgent(Robot *r,double use_cost,double speed,unsigned int index) : robot(r), index(index), use_cost(use_cost), speed(speed){
    is_human=r->getHriAgent()->is_human;
    if(is_human){shortname.append("H");}else{shortname.append("R");}
    std::string name=r->getName();
    string::iterator it=name.begin(),other_it;
    bool ok=false;
    while(!ok){
        other_it=std::find_if(it,name.end(),is_not_digit);
        if(it!=name.end() &&  other_it != name.end()){
            it=other_it;
            while(it!=name.end() && !std::isdigit(*it)){++it;}
        }else{ok=true;}
    }
    shortname.append(it,name.end());

    if(is_human){safe_dist=2;this->handover_comfort_dist=0.5;}else{safe_dist=0;}//TODO safe dists should be settable
}

double MultiHandOverAgent::reach(){
    if(robot && robot->getHriAgent() && robot->getHriAgent()->manip){
        return robot->getHriAgent()->manip->reach_tolerance;
    }
    return -numeric_limits<double>::infinity();
}

//TODO shoulderHeight is dummy constant for robots, foras long as no better solution is found
double MultiHandOverAgent::shoulderHeight(){
    Robot *r=this->robot;
    if(this->is_human){
        HRI_MANIP* manip=r->getHriAgent()->manip;
        return (*r->getJoint(manip->shoulder_joint_ids[0])->getAbsPos())[2][3];
    }else{
        return 1.25;
    }
}

MultiHandOver3dPlanner::MultiHandOver3dPlanner(Robot * robot, bool is_human) :
    _robot(robot), _is_human(is_human)
{

}



bool MultiHandOver3dPlanner::grabConfIK(configPt &q, confPtr_t const robotPos, const Eigen::Vector3d &grab_point){
    int armId = 0; // TODO select arm to use
    if(_robot->getRobotStruct()->armManipulationData->size() == 0){
        cout<<"grabConfIK : no armManipulationData for " << _robot->getName()<<endl;
        return false;
    }
    ArmManipulationData & armData = _robot->getRobotStruct()->armManipulationData->at(armId);
    ManipulationConfigs manipConf(_robot->getRobotStruct());

    //begin addition
    ManipulationPlanner * mp = new ManipulationPlanner(_robot->getRobotStruct());
    mp->setCleanningRoadmaps( false );
    setPlanningTimeLimit(60);

    fixAllJointsWithoutArm(mp->robot(),0);

    p3d_multiLocalPath_disable_all_groupToPlan(_robot->getRobotStruct(), false);
    p3d_multiLocalPath_set_groupToPlan(_robot->getRobotStruct(), mp->getUpBodyMLP(), 1, false);
    //end addition

    gpGrasp grasp;
    double confCost = -1;

    vector<vector<double> > vect;
    vector<double> target(6);

    for(unsigned int i=0;i<3;++i){
        target[i]=grab_point[i];
    }

    //target.at(3) = M_PI/2;
    //target.at(4) = (*robotPos)[11] - 7*M_PI/8 + M_PI;
    //target.at(5) = 0;

    //vect.push_back(target);

    //target.at(3) = M_PI/2;
    //target.at(4) = (*robotPos)[11] - M_PI/2 + M_PI;
    //target.at(5) = 0;

    //vect.push_back(target);

    //target.at(3) = M_PI/2;
    //target.at(4) = (*robotPos)[11] - 7*M_PI/8 + M_PI;
    //target.at(5) = P3D_HUGE;

    //vect.push_back(target);

    //target.at(3) = M_PI/2;
    //target.at(4) = P3D_HUGE;
    //target.at(5) = P3D_HUGE;

    //vect.push_back(target);

    target.at(3) = P3D_HUGE;
    target.at(4) = P3D_HUGE;
    target.at(5) = P3D_HUGE;

    vect.push_back(target);

    q=0;
    unsigned int i=0;
    //pqp_set_collision_message(1);
    while(q==0 && i<vect.size()){
        q=manipConf.getFreeHoldingConf(0,armId,grasp,armData.getCcCntrt()->Tatt, confCost, vect[i], 0);
        if(q){
            Configuration c(this->_robot,q);
            this->_robot->setAndUpdate(c);
        }
        if(q && this->_robot->isInCollisionWithOthersAndEnv()){
            q=0;
        }
        i++;
    }
    //pqp_set_collision_message(0);
    unFixAllJoints(mp->robot());
    //_robot->activateCcConstraint();
    return (q!=0);
}

bool MultiHandOver3dPlanner::grabConfIK_human(configPt &q, confPtr_t const humanPos, Eigen::Vector3d const & grab_point){
    HRI_GIK_TASK_TYPE task;
    int armId=0;
    if (armId == 0){
        task = GIK_RATREACH;
    }else{
        task = GIK_LATREACH;
    }

    p3d_vector3 Tcoord;

    Tcoord[0] = grab_point.x();
    Tcoord[1] = grab_point.y();
    Tcoord[2] = grab_point.z();
    double distance_tolerance = 0.04;
    q = p3d_get_robot_config(_robot->getRobotStruct());

    int IKSucceded = hri_agent_single_task_manip_move(_robot->getHriAgent(), task, &Tcoord, distance_tolerance, &q);
    if(IKSucceded && q){
        Configuration c(this->_robot,q);
        this->_robot->setAndUpdate(c);
    }
    if(IKSucceded && q && this->_robot->isInCollisionWithOthersAndEnv()){
        q=0;
    }
    return (q!=0);
}


bool MultiHandOver3dPlanner::grabConfCheck(configPt &q, const Eigen::Vector2d robot_pos, const Eigen::Vector3d &grab_point){
    bool ret;

    confPtr_t q_frontier(new Configuration(*_robot->getCurrentPos()));
    q_frontier->at(6+0)=robot_pos[0];
    q_frontier->at(6+1)=robot_pos[1];
    q_frontier->at(6+5)= atan2(grab_point[1]-robot_pos[1] , grab_point[0]-robot_pos[0]);
    _robot->setAndUpdate(*q_frontier);

    bool trivial_col=_robot->isInCollisionWithOthersAndEnv();

    if(trivial_col){
        ret=false;
    }else if(_is_human){
        ret = this->grabConfIK_human(q,q_frontier,grab_point);
    }else{
        ret = this->grabConfIK(q, q_frontier, grab_point);
    }
    return ret;
}

void MultiHandOver3dPlanner::clearTraj(confPtr_t conf){
    API::Trajectory p3d_trajectory(_robot);
    confPtr_t q_temp;
    if(conf){
        q_temp=confPtr_t(new Configuration(*conf));
    }else{
        q_temp=confPtr_t(new Configuration(*_robot->getCurrentPos()));
    }
    p3d_trajectory.push_back(q_temp);
    p3d_trajectory.push_back(q_temp);
    p3d_trajectory.replaceP3dTraj(_robot->getTrajStruct());
}

bool MultiHandOver3dPlanner::computeTrajectory(confPtr_t start_conf,confPtr_t nav_conf, std::vector<Eigen::Vector2d,Eigen::aligned_allocator<Eigen::Vector2d> > const& path2d, confPtr_t end_conf)
{
    API::Trajectory p3d_trajectory(_robot);
    confPtr_t q_save(new Configuration(*_robot->getCurrentPos()));

    if(path2d.size()>1){

        std::vector<Eigen::Vector3d, Eigen::aligned_allocator<Eigen::Vector3d> > path3d;
        this->computePath2dToPath3dShortCut(path2d,path3d);

        //go from start to nav
        //(re)move virtual object
//        if(!_is_human){
//            unsigned int f_voi_dof=_robot->getJoint("virtual_object_right")->getIndexOfFirstDof();
//            for(unsigned int voi=f_voi_dof;voi<f_voi_dof+6;voi++){
//                (*start_conf)[voi] = 0;
//            }
//        }
        (*start_conf)[6+0]=path3d[0][0];//x
        (*start_conf)[6+1]=path3d[0][1];//y
        //(*start_conf)[6+5]=angle_limit_PI(path3d[0][2]);//theta
        p3d_trajectory.push_back(start_conf);

        for (unsigned int i=0;i<path3d.size();++i){
            confPtr_t q_temp(new Configuration(*nav_conf));
            (*q_temp)[6+0]=path3d[i][0];//x
            (*q_temp)[6+1]=path3d[i][1];//y
            (*q_temp)[6+5]=angle_limit_PI(path3d[i][2]);//theta
            if(i==0){
                //the first theta is dummy, set it as the starting pos theta
                (*q_temp)[6+5]=angle_limit_PI((*start_conf)[6+5]);//theta
            }

            p3d_trajectory.push_back(q_temp);
            //attempt to make the robot stop before doing handover
            //if(i==path3d.size()-1){
            //    confPtr_t q_temp2(new Configuration(*q_temp));
            //    (*q_temp2)[6+5] = (*end_conf)[6+5];
            //    p3d_trajectory.push_back(q_temp2);
            //}
        }
        if(end_conf){
            //attempt to make the robot stop before doing handover
            //the ho 2d position with previous theta
            confPtr_t q_temp1(new Configuration(*nav_conf));
            (*q_temp1)[6+0] = (*end_conf)[6+0];//x
            (*q_temp1)[6+1] = (*end_conf)[6+1];//y
            if(path2d.size()==2){
                //the path  is no move, the end theta is dummy as the first (0)
                //set it to be = to the first
                (*q_temp1)[6+5]=angle_limit_PI((*start_conf)[6+5]);//theta
            }else{
                (*q_temp1)[6+5] = angle_limit_PI(path3d.back()[2]);//theta
            }
            p3d_trajectory.push_back(q_temp1);
            //the ho 2d position with ho theta
            confPtr_t q_temp2(new Configuration(*q_temp1));
            (*q_temp2)[6+5] = (*end_conf)[6+5];//theta
            p3d_trajectory.push_back(q_temp2);
            //the full 3d ho conf
            //move virtual object
            //move virtual object
            // if(!_is_human){
            //    unsigned int f_voi_dof=_robot->getJoint("virtual_object_right")->getIndexOfFirstDof();
            //      for(unsigned int voi=f_voi_dof;voi<f_voi_dof+6;voi++){
            //          (*end_conf)[voi] = 0;
            //      }
            //  }
            p3d_trajectory.push_back(end_conf);
        }
    }else{
        cout << "not computing trajectory, path contains only "<<path2d.size()<< " points"<<endl;
    }


    bool  trajTest;
    trajTest = p3d_trajectory.size() != 0;
    if(trajTest){
        trajTest = p3d_trajectory.replaceP3dTraj(_robot->getTrajStruct());
    }else{
        cout << "computed a traj of null size"<<endl;
    }
    if(trajTest){
        cout << "traj ok"<<endl;
    }else{
        cout << "traj error"<<endl;
    }

    return trajTest;

}

bool MultiHandOver3dPlanner::computePath2dToPath3dShortCut(
        std::vector<Eigen::Vector2d, Eigen::aligned_allocator<Eigen::Vector2d> > const& path2d,
        std::vector<Eigen::Vector3d, Eigen::aligned_allocator<Eigen::Vector3d> > &path3d)
{
    PlannarTrajectorySmoothing *m_pts=new PlannarTrajectorySmoothing(_robot);
    path3d.clear();
    if(path2d.size()>1){
        //smooth and convert to 3d path
        path3d = m_pts->add3Dim(m_pts->smoothTrajectory(_robot,path2d),_robot,0.05);
    }else if(path2d.size()==1){
        path3d.push_back(Eigen::Vector3d(path2d.front().x(),path2d.front().y(),0));//TODO computePath2dToPath3dShortCut path.size==1 => theta is dummy
    }
    return path3d.size() !=0;
}

bool MultiHandOver3dPlanner::computeHandOverTrajectory(confPtr_t q_start, confPtr_t q_end){
    //update confs
    _robot->setGoTo(*q_end);
    _robot->setAndUpdate(*q_start);

    global_Project->getActiveScene()->setActiveRobot(_robot->getName());
    API::Trajectory p3d_trajectory(_robot);

    p3d_trajectory.push_back(q_start);
    p3d_trajectory.push_back(q_end);

    bool trajTest = p3d_trajectory.replaceP3dTraj(_robot->getTrajStruct());
    if(trajTest){
        cout << "traj ok"<<endl;
    }else{
        cout << "traj error"<<endl;
    }

    return trajTest;
}

/*
bool MultiHandOver3dPlanner::computeHandOverTrajectory(confPtr_t q_start, confPtr_t q_end){
    //update confs
    _robot->setGoTo(*q_end);
    _robot->setAndUpdate(*q_start);

    global_Project->getActiveScene()->setActiveRobot(_robot->getName());

    try
    {
#ifdef P3D_PLANNER
        p3d_SetStopValue(FALSE);
#endif
        timeval tim;
        gettimeofday(&tim, NULL);
        double t_init = tim.tv_sec+(tim.tv_usec/1000000.0);

        cout << "ENV.getBool(Env::Env::treePlannerIsEST) = " << ENV.getBool(Env::treePlannerIsEST) << endl;
        if (ENV.getBool(Env::treePlannerIsEST))
        {
            p3d_run_est(_robot->getRobotStruct());
        }
        else
        {
            p3d_run_rrt(_robot->getRobotStruct());
        }

        gettimeofday(&tim, NULL);
        double dt = tim.tv_sec+(tim.tv_usec/1000000.0) - t_init;
        cout << "RRT computed in : " << dt << " sec" << endl;

        if( !ENV.getBool(Env::drawDisabled) ) {
            g3d_draw_allwin_active();
        }
    }
    catch (string str)
    {
        cerr << "Exeption in run computeHandOverTrajectory() : " << endl;
        cerr << str << endl;
        ENV.setBool(Env::isRunning,false);
    }
    catch (...)
    {
        cerr << "Exeption in run computeHandOverTrajectory()" << endl;
        ENV.setBool(Env::isRunning,false);
    }
    return true;
}
*/

MultiHandOverSolution::CostData::CostData(MultiHandOverSolution *sol){
    handovers=sol->actions.size();
    dist=sol->getTotalDist();
    time=sol->time;
    use_cost=sol->getTotalUseCost();
    handover_cost=sol->hri_cost;
    marker_is_opt=sol->is_optimized;
    go_to_target_nb=sol->navigation_trajs_to_target.size();
    for(unsigned int i=0;i<sol->agents.size();++i){
        this->agent_names.push_back(sol->agents[i]->robot->getName());
    }
}

unsigned int MultiHandOverSolution::_current_id=0;

MultiHandOverSolution::MultiHandOverSolution() : cost(-1), is_optimized(0), sons(0), parent(0), oldest_parent(0),
     _id(_current_id++)
{ }

MultiHandOverSolution * MultiHandOverSolution::cloneAsSon(){
    MultiHandOverSolution * son=this->clone();
    son->_id=_current_id++;
    this->sons.push_back(son);
    son->parent=this;
    if(this->oldest_parent){son->oldest_parent=this->oldest_parent;}
    else{son->oldest_parent=this;}
    son->sons.clear();
    return son;
}

MultiHandOverSolution * MultiHandOverSolution::clone(){
    MultiHandOverSolution * clone=new MultiHandOverSolution;
    *clone=*this;
    clone->actions.clear();
    for (unsigned int i=0; i<clone->frontiers.size();i++){
        clone->actions.push_back(Action(i,clone));
    }
    return clone;
}

bool MultiHandOverSolution::addHandOverAction(FrontierPtr frontier){
    unsigned int i=this->frontiers.size();//where it will be added
    //check order of frontiers and agents is conform :
    if( i+1>=this->agents.size()
            || ((frontier->agentA->robot() == this->agents[i]->robot && frontier->agentB->robot() != this->agents[i+1]->robot)
                || (frontier->agentB->robot() == this->agents[i]->robot && frontier->agentA->robot() != this->agents[i+1]->robot ) )
            ){
        return false;
    }
    this->frontiers.push_back(frontier);
    this->actions.push_back(Action(i,this));
    return true;
}

double MultiHandOverSolution::getTotalDist(){
    double d=0;
    for (unsigned int i=0;i<this->nav_dists.size();++i){
        d+=this->nav_dists[i];
    }
    return d;
}
double MultiHandOverSolution::getTotalUseCost(){
    double c=0;
    for (unsigned int i=0;i<this->actions.size();++i){
        Action a=this->actions[i];
        c+=a.dist(0)/a.agent(0)->speed*a.agent(0)->use_cost;
        c+=a.dist(1)/a.agent(1)->speed*a.agent(1)->use_cost;
    }
    return c;
}

MultiHandOverSolution::Action::Action(unsigned int frontier_index, MultiHandOverSolution * solution):
    _f(frontier_index), _s(solution)
{
}

confPtr_t MultiHandOverSolution::Action::agentConf(unsigned int i){
    checkI(i);

    if(this->frontier()->isSymetric() || this->robot(0)==this->frontier()->agentA->robot()){
        //symetric or direct
        if(i==0)
            return this->frontier()->confA;
        else
            return this->frontier()->confB;
    }else{
        if(i==0)
            return this->frontier()->r_confB;
        else
            return this->frontier()->r_confA;
    }
}

confPtr_t MultiHandOverSolution::Action::objectPos(){
    if(this->frontier()->isSymetric() || this->robot(0) == this->frontier()->agentA->robot() ){
        //symetric or direct
        return this->frontier()->object_position;
    }else{
        return this->frontier()->r_object_position;
    }
}

string MultiHandOverSolution::print_long(){
    stringstream ss;
    ss<< "id: "<<getId();
    if(this->parent){ss<<" son of: "<<this->parent->getId();}
    ss << endl;
    ss<<"agents: ";
    for (unsigned int i=0;i<agents.size();i++){
        ss<<agents[i]->robot->getName()<<" - ";
    }
    ss<<endl<<"COST = "<<cost<<endl;
    ss<<"TIME = "<<this->time<<endl;
    if(this->is_optimized){
        ss<<"-> is optimized\n";
    }
    ss<<"frontiers :"<<endl;
    for (unsigned int i=0;i<this->frontiers.size();i++){
        ss << '\t'<<i<<"- CS:"<<frontiers[i]->check_status<<" o:"<<frontiers[i]->status_ok[FrontierCells::FCS_OBJECT]<<" g:"<<frontiers[i]->status_ok[FrontierCells::FCS_GRAB_CONFS]
           << " (gA:"<<(frontiers[i]->confA ? 1: 0)<<" gB:"<<(frontiers[i]->confB ? 1 : 0)<<")"<< endl;
    }
    ss<<"nav dists :"<<endl;
    for (unsigned int i=0;i<this->nav_dists.size();i++){
        ss<<i<<": "<<nav_dists[i]<<" - ";
    }
    ss<< endl<<"ho times :"<<endl;
    for (unsigned int i=0;i<this->handover_times.size();i++){
        ss<<i<<": "<<handover_times[i]<<" - ";
    }
    ss<<endl<<"hri cost : "<<this->hri_cost;
    ss<<endl<<"nav to targets dists :\n";
    for (unsigned int i=0;i<this->nav_dists_to_target.size();++i){
        ss<<i<<": "<<this->nav_dists_to_target[i]<<" - ";
    }
    ss<<endl<<"--"<<endl<<endl;
    return ss.str();
}

std::ostream& operator<<(std::ostream& os, const MultiHandOverSolution& sol)
{
    os<<"cost: "<<sol.cost<<" ";
    for(unsigned int i=0;i<sol.agents.size();++i){
        os<<" - "<<sol.agents[i]->robot->getName();
    }
    return os;
}

bool operator<(const MultiHandOverSolution &left, const MultiHandOverSolution &right){
    return left.cost < right.cost;
}
bool operator>(const MultiHandOverSolution &left, const MultiHandOverSolution &right){
    return left.cost > right.cost;
}
bool MultiHandOverSolution::less(MultiHandOverSolution const *left,MultiHandOverSolution const *right){
    return *left<*right;
}

void MultiHandOverSolution::printFamiliesTrees(std::vector<MultiHandOverSolution *> &solutions, ofstream &output){
    int red[3]={0x00,0x00,0x88};
    int green[3]={0x88,0xff,0x88};
    int color_tab[3];
    int color;
    int cumulated=0;
    double ratio;
    output<< "digraph MHOSolutions {"<<endl << "\tnode[style=filled]"<<endl;
    output<< "\tlegend[label=\"ID\\ncost(index)\\nagent sequence\",style=\"dashed\"];"<<endl
          << "\tlegend_son[label=\"ID\\ncost(index)\\n(same sequence)\",style=\"dashed\"];"<<endl
          << "\tlegend->legend_son;"<<endl;
    if(solutions.size()==0){
        output<<'\t'<<"no_solutions[style=\"dashed\", color=\"black\",label=\"no solutions\"];"<<endl;
    }else{
        double worst=solutions.back()->cost;
        double best=solutions.front()->cost;
        for (unsigned int i=0;i<solutions.size();++i){
            if(i>solutions.size()/3 && solutions[i]->parent==0 && solutions[i]->sons.size()==0){
                //bad solution and orphan, not interesting
                cumulated++;
            }else{
                if(cumulated>0){
                    //have cumulated not interesting nodes, print it
                    output<<'\t'<<"cumul"<<i<<"[style=\"dashed\", color=\"black\",label=\"cumulated "<<cumulated<<"solutions\"];"<<endl;
                    cumulated=0;
                }
                ratio=(solutions[i]->cost-best)/(worst-best);
                ratio=ratio*5;
                if(ratio>1){ratio=1;}
                color=0;
                for(int ic=0;ic<3;++ic){
                    color_tab[ic]=ratio*red[ic] + (1-ratio)*green[ic];
                    color += color_tab[ic]<<8*ic;
                }
                output<<'\t'<<solutions[i]->getId()<<"[color=\"#";
                output << std::hex<<std::setw(6)<<std::setfill('0')<<color<<std::dec<<'\"';
                if(i==0){
                    output << ", style=bold";
                }
                output<<",label=\"\\N\\n"<<solutions[i]->cost<<'('<<i<<')';
                //add agent names if this is the oldest of the family
                if(solutions[i]->parent==0){
                    output<<"\\n";
                    for (unsigned int j=0;j<solutions[i]->agents.size();++j){
                        output<< solutions[i]->agents[j]->shortname<<',';
                    }
                }
                output<<"\"";//close label
                output << "];"<<endl<<'\t';//close node parameter, end line
                if(solutions[i]->parent){
                    output << solutions[i]->parent->getId()<<"->";
                }
                output << solutions[i]->getId()<<";"<<endl;
            }
        }
        if(cumulated>0){
            //have cumulated not interesting nodes, print it
            output<<'\t'<<"cumul_end[style=\"dashed\", color=\"black\",label=\"cumulated "<<cumulated<<"solutions\"];"<<endl;
            cumulated=0;
        }
    }
    output<<'}'<<endl;
    output.flush();
}


// ////////////////////////////////////////////////////
// ///////////        MHO [SUB] TASK      /////////////
// ////////////////////////////////////////////////////

unsigned int SubTask::_count_id=0;

SubTask::SubTask():
    _type(UNDEF),
    _handoverStepType(SHO_NONE),
    _begin_date(0),_end_date(0),_duration(0),_cost(0),
    _agent(0),
    _wp(0),_traj(0),
    _id(_count_id++)

{
    resetMarker();
}
SubTask::SubTask(MultiHandOverAgent *agent, TaskType_t type, double beginDate, double endDate,double cost, HandoverStepType_t hostepType):
    _type(type),
    _handoverStepType(hostepType),
    _begin_date(beginDate),_end_date(endDate),_duration(endDate-beginDate),_cost(cost),
    _agent(agent),
    _wp(0),_traj(0),
    _id(_count_id++)
{
    //assert mandatory stuff (move to exception?)
    assert(_agent && _duration>=0);
    assert(_begin_date >=0 && _end_date >=0 && _end_date >= _begin_date);
    resetMarker();
}
SubTask::~SubTask(){}
void SubTask::resetMarker(){
    this->_markers.assign(SIZE,false);
}
bool SubTask::getMarker(MarkerList_t m){return _markers[m];}
void SubTask::setMarker(MarkerList_t m, bool b){_markers[m]=b;}

SubTask::TaskType_t SubTask::type() const {return _type;}
SubTask::HandoverStepType_t SubTask::hoStepType() const {return _handoverStepType;}
bool SubTask::isNav() const {
    return (_type==NAV_OTHER || _type==NAV_TO_HO || _type==NAV_TO_TARGET || _type==NAV_WITH_OBJ || _type==NAV_SYNC);
}
bool SubTask::isManip() const {
    return (_type==HO_GET || _type==HO_GIVE);
}

double SubTask::beginDate(){return _begin_date;}
double SubTask::endDate(){return _end_date;}
double SubTask::duration(){return _duration;}
void SubTask::extendToEndDate(double new_end_date){_end_date=new_end_date;_duration=_end_date-_begin_date;}
void SubTask::extendDuration(double new_duration){_duration=new_duration;_end_date=_begin_date+_duration;}
void SubTask::replan(double new_begin){
    if(new_begin < -10){
        //TODO debug
        return;
    }
    double begin_bkp(_begin_date),end_bkp(_end_date),delta=new_begin-begin_bkp;
    if(abs(new_begin-begin_bkp) < 0.000001 ){
        //no change
        return;
    }
    _begin_date=new_begin;
    _end_date=_begin_date+_duration;
    if(type()==NAV_SYNC){
        for(unsigned int i=0;i<_dates_wp.size();i++){
            _dates_wp[i]+=delta;
        }
    }
    //report next tasks of this
    for(vector<SubTask*>::iterator it = this->next.begin() ; it!=this->next.end();++it){
        SubTask *s=*it;
        double va=std::fabs(s->beginDate() - end_bkp);
        if ( s->type() != IDLE && va <= 0.000001 ){
            //the sub task has to be changed, its begin date coincide with the old end date of this
            s->replan(_end_date);
        }
        //assert the begin date of the other is superior to the end date of this one
        //assert(s->beginDate() >= _end_date);
    }

    //report prev tasks of this as well
    for(vector<SubTask*>::iterator it = this->prev.begin() ; it!=this->prev.end();++it){
        SubTask *s=*it;
        if ( s->type() != IDLE && abs(s->endDate() - begin_bkp) <= 0.000001 ){
            //the sub task has to be changed, its begin date coincide with the old end date of this
            s->replan(_begin_date - s->duration());
        }
        //assert the begin date of the other is superior to the end date of this one
        //assert(s->endDate() <= _begin_date);
    }
}

double SubTask::cost(){return _cost;}
MultiHandOverAgent* SubTask::agent(){return _agent;}
std::vector<confPtr_t> & SubTask::waypoints(){return _wp;}
std::vector<Eigen::Vector2d> & SubTask::waypoints2d(){return _2d_wp;}
void SubTask::addWaypoint(confPtr_t q){
    this->_wp.push_back(q);
}
void SubTask::addWaypoint(Eigen::Vector2d v,double date){
    _2d_wp.push_back(v);
    if(type()==NAV_SYNC){
        _dates_wp.push_back(date);
        assert(date!=-numeric_limits<double>::infinity());
    }
}
vector<double> & SubTask::waypointDates(){return _dates_wp;}
API::Trajectory & SubTask::traj(){return _traj;}
void SubTask::traj(API::Trajectory traj){_traj=traj;}

void SubTask::spreadIdlePositionToNext(){
    for(vector<SubTask*>::iterator it = this->next.begin() ; it!=this->next.end();++it){
        if((*it)->type()==IDLE && (*it)->waypoints2d().front() != this->waypoints2d().front()){
            (*it)->waypoints2d().front()=this->waypoints2d().front();
            (*it)->spreadIdlePositionToNext();
        }
    }
}

void SubTask::addNext(SubTask *next){
    if(next)
        assert(next->type()!=NAV_TO_HO || next->agent() == this->agent());
        this->next.push_back(next);
}
void SubTask::addPrevious(SubTask *prev){
    if(prev)
        this->prev.push_back(prev);
}

void SubTask::substituteNext(SubTask *next, SubTask *new_next){
    if(new_next){
        std::replace(this->next.begin(),this->next.end(),next,new_next);
    }else{
        this->next.erase(std::find(this->next.begin(),this->next.end(),next));
    }
}
void SubTask::substitutePrev(SubTask *prev, SubTask *new_prev){
    if(new_prev){
        std::replace(this->prev.begin(),this->prev.end(),prev,new_prev);
    }else{
        this->prev.erase(std::find(this->prev.begin(),this->prev.end(),prev));
    }
}


Task::Task(MultiHandOverSolution *sol){
    vector<SubTask*> aux_tasks(0);
    _agents=sol->agents;
    this->setObject(sol->object);
    for(unsigned int i=0;i<sol->actions.size();++i){
        MultiHandOverSolution::Action* act=&sol->actions[i];
        FrontierPtr f=act->frontier();
        vector<SubTask*> prev_tasks=aux_tasks;
        aux_tasks.clear();
        //add the navigation tasks
        for (unsigned int a=0;a<2;a++){
            SubTask::TaskType_t type;
            if(a==0){type=SubTask::NAV_WITH_OBJ;}
            else{type=SubTask::NAV_TO_HO;}
            double beg_date=act->startDate(a);
            SubTask* tk=new SubTask(act->agent(a),type,beg_date,beg_date+act->navTime(a));
            tk->comment=string("nav from sol");
            tk->setMarker(SubTask::IS_2D,true);
            //add the path as 2d waypoints
            confPtr_t qprev;
            if( type == SubTask::NAV_WITH_OBJ && i >0){
                MultiHandOverSolution::Action *prev_act = &sol->actions[i-1];
                qprev=prev_act->frontier()->confOf(act->robot(a));
                Eigen::Vector2d prev_hopos;
                prev_hopos.x() = qprev->at(6+0);
                prev_hopos.y() = qprev->at(6+1);
                //add the first 2d wp as the exact position of the previous handover
                tk->addWaypoint(prev_hopos);
            }
            for (unsigned int w=0;w<act->path(a).size();++w){
                MultiAgentCell* cell=act->path(a)[w];
                tk->addWaypoint(cell->getCenter());
            }
            Eigen::Vector2d hopos;
            confPtr_t qho = confPtr_t(new Configuration(*act->frontier()->confOf(act->robot(a))));
            confPtr_t qtarget = confPtr_t(new Configuration(act->initQ(a)));
            confPtr_t qbegin = confPtr_t(new Configuration(act->initQ(a)));
            for(unsigned int qi=6;qi<6+6;qi++){
                qtarget->at(qi) = qho->at(qi);
                if(type == SubTask::NAV_WITH_OBJ && i >0){
                    //begin is previous handover
                    qbegin->at(qi) = qprev->at(qi);
                    assert(qbegin->getRobot() == qprev->getRobot());
                }
            }
            tk->addWaypoint(qbegin);
            tk->addWaypoint(qtarget);
            //insert exact ho pos after
            hopos.x() = qtarget->at(6+0);
            hopos.y() = qtarget->at(6+1);
            tk->addWaypoint(hopos);
            if(tk->waypoints2d().size()==1){
                tk->addWaypoint(tk->waypoints2d().back());
            }
            if(type==SubTask::NAV_WITH_OBJ){
                //only the navigation of the object holder have predecessors
                //set the previous tasks of this
                tk->getPrev()=prev_tasks;
                for(unsigned int p=0;p<prev_tasks.size();p++){
                    prev_tasks[p]->addNext(tk);
                }
            }
            //add to the auxiliary to be added as previous tasks for the hand over tasks
            aux_tasks.push_back(tk);
            //add to this
            this->_sub_tasks.push_back(tk);
        }

        //add the hand over tasks
        prev_tasks=aux_tasks;
        aux_tasks.clear();
        for (unsigned int a=0;a<2;a++){
            SubTask::TaskType_t type;
            if(a==0){type=SubTask::HO_GIVE;}
            else{type=SubTask::HO_GET;}
            SubTask* tk=new SubTask(act->agent(a),type,act->startDateHO(),act->endDateHO());
            tk->comment=string("HO from sol");

            //add the key positions as configuration (Q) waypoints
            confPtr_t qtarget=confPtr_t(new Configuration(*f->confOf(act->robot(a))));
            confPtr_t qinit=confPtr_t(new Configuration(*act->robot(a)->getInitialPosition()));
            for(unsigned int i=0;i<6;i++){
                qinit->at(6+i) = qtarget->at(6+i);
            }
            tk->addWaypoint(qinit);
            tk->addWaypoint(qtarget);
            Eigen::Vector2d hopos;
            hopos.x() = qtarget->at(6+0);
            hopos.y() = qtarget->at(6+1);
            tk->addWaypoint(hopos);

            //set the previous tasks of this (the 2 navigations)
            tk->getPrev()=prev_tasks;
            for(unsigned int p=0;p<prev_tasks.size();p++){
                prev_tasks[p]->addNext(tk);
            }
            //add to the auxiliary list to be set as previous tasks of
            //the navigation task of the new holder
            aux_tasks.push_back(tk);
            //add to this
            this->_sub_tasks.push_back(tk);
        }
        //add the final navigations
        if(sol->navigation_trajs_to_target[i].size() && sol->nav_dists_to_target[i]>0){
            std::vector<HRICS::MultiAgentCell*> &path=sol->navigation_trajs_to_target[i];
            SubTask* tk=new SubTask(sol->agents[i],SubTask::NAV_TO_TARGET,sol->nav_start_end_times_to_target[i].first,sol->nav_start_end_times_to_target[i].second);
            tk->comment=string("final of interm from sol");
            tk->setMarker(SubTask::IS_2D,true);

            SubTask *ptk(0);
            for(unsigned int p=0;p<aux_tasks.size(); p++){
                SubTask *ppp=ptk=aux_tasks[p];
                if(ppp->agent() == tk->agent() && ppp->type() == SubTask::HO_GIVE){
                    ptk=ppp;
                    break;
                }
            }
            assert(ptk);
            tk->addWaypoint(confPtr_t(new Configuration(*ptk->waypoints().front())));
            tk->addWaypoint(confPtr_t(new Configuration(sol->target_confs[i])));
            tk->addWaypoint(ptk->waypoints2d().back());

            //set the previous tasks of this (the 2 tasks related to the hand over)
            tk->getPrev()=aux_tasks;
            for(unsigned int p=0;p<aux_tasks.size();p++){
                aux_tasks[p]->addNext(tk);
            }
            //add the path as 2d waypoints
            for (unsigned int w=0;w<path.size();++w){
                MultiAgentCell* cell=path[w];
                tk->addWaypoint(cell->getCenter());
            }
            this->_sub_tasks.push_back(tk);
        }

    }
    //add the final navigation of the last agent
    if(sol->navigation_trajs_to_target.back().size()){
            std::vector<HRICS::MultiAgentCell*> &path=sol->navigation_trajs_to_target.back();
            SubTask* tk=new SubTask(sol->agents.back(),SubTask::NAV_WITH_OBJ,sol->nav_start_end_times_to_target.back().first,sol->nav_start_end_times_to_target.back().second);
            tk->comment=string("final of last from sol");
            tk->setMarker(SubTask::IS_2D,true);

            SubTask *ptk(0);
            for(unsigned int p=0;p<aux_tasks.size(); p++){
                SubTask *ppp=ptk=aux_tasks[p];
                if(ppp->agent() == tk->agent() && ppp->type() == SubTask::HO_GIVE){
                    ptk=ppp;
                    break;
                }
            }
            assert(ptk);
            tk->addWaypoint(ptk->waypoints().front());
            tk->addWaypoint(ptk->waypoints2d().back());
            tk->addWaypoint(confPtr_t(new Configuration(sol->target_confs.back())));

            //add the path as 2d waypoints
            for (unsigned int w=0;w<path.size();++w){
                MultiAgentCell* cell=path[w];
                tk->addWaypoint(cell->getCenter());
            }
            tk->addPrevious(aux_tasks.back());
            aux_tasks.back()->addNext(tk);
            this->_sub_tasks.push_back(tk);
    }
    this->refineHandOvers();
}

Task::~Task(){
    for(unsigned int i=0;i<_sub_tasks.size();i++){
        delete _sub_tasks[i];
    }
}

void Task::refineHandOvers(){
    SubTask *aux(0);
    for (unsigned int i=0;i<_sub_tasks.size();i++){
        if (    getSubTask(i)->hoStepType() == SubTask::SHO_NONE
                && (    getSubTask(i)->type() == SubTask::HO_GET
                     || getSubTask(i)->type() == SubTask::HO_GIVE)
                ){
            if(aux){
                refineHandOver(aux,getSubTask(i));
                aux=0;
            }else{
                aux=getSubTask(i);
            }
        }
    }
}

void Task::refineHandOver(SubTask *st1, SubTask *st2){
    if(st1->type() == SubTask::HO_GET){
        //st1 must be give
        std::swap(st1,st2);
    }
    assert(st1->type()==SubTask::HO_GIVE && st2->type() == SubTask::HO_GET);
    assert(st1->hoStepType() == SubTask::SHO_NONE && st2->hoStepType() == SubTask::SHO_NONE);

    SubTask *give,*pick,*release,*grab,*diseng,*get; // the six steps of the handover
    double bt1 = st1->beginDate();
    double et1 = bt1 + st1->duration()/3;
    double bt2 = et1;
    double et2 = bt2 + st2->duration()/3;
    give    = new SubTask(st1->agent(),SubTask::HO_GIVE,bt1,et1,0,SubTask::SHO_G_GIVE);
    pick    = new SubTask(st2->agent(),SubTask::HO_GET,bt2,et2,0,SubTask::SHO_R_PICK);
    bt1=et2;
    et1=et2;
    bt2=et2;
    release = new SubTask(st1->agent(),SubTask::HO_GIVE,bt1,et1,0,SubTask::SHO_G_RELEASE);
    grab    = new SubTask(st2->agent(),SubTask::HO_GET,bt2,et2,0,SubTask::SHO_R_GRAB);
    et1+=st1->duration()/3;
    et2+=st2->duration()/3;
    diseng  = new SubTask(st1->agent(),SubTask::HO_GIVE,bt1,et1,0,SubTask::SHO_G_DISENGAGE);
    get     = new SubTask(st2->agent(),SubTask::HO_GET,bt2,et2,0,SubTask::SHO_R_TAKE);
    std::vector<SubTask*> news;
    news.push_back(give);
    news.push_back(pick);
    news.push_back(grab);
    news.push_back(release);
    news.push_back(get);
    news.push_back(diseng);


    //make prev/next links
    for(unsigned int i=0;i<st1->getPrev().size();i++){
        give->addPrevious(st1->getPrev()[i]);
        st1->getPrev()[i]->substituteNext(st1,give);
        st1->getPrev()[i]->substituteNext(st2,0);
    }
    give->getPrev()=st1->getPrev();

    give->addNext(pick);
    pick->addPrevious(give);
    pick->addNext(grab);
    grab->addPrevious(pick);
    grab->addNext(release);
    release->addPrevious(grab);
    release->addNext(get);
    release->addNext(diseng);
    get->addPrevious(release);
    diseng->addPrevious(release);

    diseng->getNext()=st1->getNext();
    get->getNext()=st2->getNext();
    for(unsigned int i=0;i<st1->getNext().size();i++){
        st1->getNext()[i]->substitutePrev(st1,diseng);
    }
    for(unsigned int i=0;i<st2->getNext().size();i++){
        st2->getNext()[i]->substitutePrev(st2,get);
    }

    //add waypoints
    give->waypoints() = st1->waypoints();
    give->waypoints2d() = st1->waypoints2d();

    pick->waypoints() = st2->waypoints();
    pick->waypoints2d() = st2->waypoints2d();

    release->addWaypoint(give->waypoints().back());
    release->addWaypoint(give->waypoints2d().back());

    grab->addWaypoint(pick->waypoints().back());
    grab->addWaypoint(pick->waypoints2d().back());

    diseng->waypoints().assign(st1->waypoints().rbegin(),st1->waypoints().rend());
    diseng->waypoints2d().assign(st1->waypoints2d().rbegin(),st1->waypoints2d().rend());

    get->waypoints().assign(st2->waypoints().rbegin(),st2->waypoints().rend());
    get->waypoints2d().assign(st2->waypoints2d().rbegin(),st2->waypoints2d().rend());

    vector<SubTask*>::iterator it=std::find(_sub_tasks.begin(),_sub_tasks.end(),st1);
    this->_sub_tasks.insert(it,news.begin(),news.end());
    it=std::find(_sub_tasks.begin(),_sub_tasks.end(),st1);
    this->_sub_tasks.erase(it);
    this->_sub_tasks.erase(std::find(_sub_tasks.begin(),_sub_tasks.end(),st2));
    delete st1;
    delete st2;

}

SubTask * Task::getSubTask(int i){
    if(i>=0 && (unsigned int)i<_sub_tasks.size()){
        return _sub_tasks[i];
    }else
        return 0;
}
const std::vector<SubTask *> &Task::getSubTasks(){
    return this->_sub_tasks;
}

std::string SubTask::typeString(SubTask::TaskType_t t){
    std::string type;
    switch(t){
    case SubTask::NAV_TO_HO:
    case SubTask::NAV_WITH_OBJ:
        type.assign("na");
        break;
    case SubTask::HO_GET:
    case SubTask::HO_GIVE:
        type.assign("ho");
        break;
    case SubTask::NAV_TO_TARGET:
        type.assign("tg");
        break;
    case SubTask::IDLE:
        type.assign("zz");
        break;
    default:
        type.assign("?");
        break;
    }
    return type;
}
void Task::plotGanttChart(){
    char T='\t';
    ofstream of;
    of.open("task_gantt.dat");
    of<<"#Task start      end"<<endl;
    for(unsigned int i=0;i<this->_sub_tasks.size();i++){
        std::string type = SubTask::typeString(this->getSubTask(i)->type());
        if(this->getSubTask(i)->type()!=SubTask::IDLE){
            of << this->getSubTask(i)->agent()->shortname
               << '-'<<type //<< '('<<this->getSubTask(i)->getId()<<')'
               << T<<this->getSubTask(i)->beginDate()<<T<<this->getSubTask(i)->endDate()<<endl;
        }
    }
    of.close();
    of.open("task_gantt.plot");
    of <<
       "set term wxt persist\n"
    << "# set terminal svg size 600,400 dynamic enhanced fname 'arial'  fsize 10 mousing name \"gantt_1\" butt dashlength 1.0\n"
    << "# set output 'gantt.1.svg'\n"
    << "set border 3 front lt rgb \"black\" linewidth 1.000\n"
    << "set grid nopolar\n"
    << "set grid xtics nomxtics ytics nomytics noztics nomztics \\\n"
    << " nox2tics nomx2tics noy2tics nomy2tics nocbtics nomcbtics\n"
    << "set grid layerdefault   lt rgb \"black\" linewidth 0.700,  lt rgb \"black\" linewidth 0.700\n"
    << "unset key\n"
    << "set style arrow 1 head back filled linecolor rgb \"#56b4e9\"  linewidth 1.500 size screen  0.020,15.000,90.000  #fixed\n"
    << "set style data lines\n"
    << "set mxtics 4.000000\n"
    << "set xtics border in scale 2,0.5 nomirror norotate  autojustify\n"
    << "#set xtics 2.6784e+06 norangelimit\n"
    << "set ytics border in scale 1,0.5 nomirror norotate  autojustify\n"
    << "set ytics  norangelimit\n"
    << "set ytics   ()\n"
    << "#set title \"{/=15 Simple Gantt Chart}\\n\\n{/:Bold Task start and end times in columns 2 and 3}\"\n"
    << "set yrange [ -1.00000 : * ] noreverse nowriteback\n"
    << "set xrange [ -5.00000 : * ] \n"
    << "x = 0.0\n"
    << "plot \"task_gantt.dat\" using ($2) : ($0) : (($3)-($2)) : (0.0) : yticlabel(1) with vector as 1, \\\n"
    << "    \"task_gantt.dat\" using ($2) : ($0) : 1 with labels right offset -2\n"
    << endl;
    of.close();
    system("gnuplot task_gantt.plot");
}

void Task::displayGraph(){
    ofstream of;
    char T='\t', g='\"';
    of.open("task_time_graph.dot");
    of << "digraph task_time_graph {"<<endl
       << T << "rankdir=LR;"<<endl;
    for(unsigned int i=0;i<_sub_tasks.size();i++){
        SubTask* st=_sub_tasks[i];
        std::string type = SubTask::typeString(this->getSubTask(i)->type());
        std::string  supp;
        if(st->type()==SubTask::IDLE){
            supp=", style=\"filled\", color=\"lightblue\"";
        }else if(st->type()==SubTask::HO_GET || st->type()==SubTask::HO_GIVE){
            supp=", style=\"filled\", color=\"orange\"";
        }
        of << T
           << g << st->getId() << g
              //<< " [ label = \""<< i << "-" << st->agent()->shortname
              << " [ label = \""<< st->agent()->shortname
              << "-" << type << "\"" << supp <<  "];" << endl;
        for(unsigned int p=0;p<st->getPrev().size();p++){
            SubTask *pst=st->getPrev()[p];
            std::string supp_ed="";
            if(pst->endDate()>st->beginDate()){
                supp_ed+=", color=\"red\"";
            }
            if( pst->agent() == st->agent() && pst->waypoints2d().back() != st->waypoints2d().front()){
                //same agent but not the sames start/end poses
                supp_ed+=", color=\"purple\"";
            }
            if( pst->agent() == st->agent() && pst->waypoints().size() && st->waypoints().size()
                    && !pst->waypoints().back()->equal(*st->waypoints().front())){
                //same agent but not the sames start/end poses
                supp_ed+=", penwidth=5";
            }
            of << T;
            of << g << pst->getId() << g
               << "->" << g << st->getId() << g
               << "[ label = \"" << pst->duration();
            if( abs(st->beginDate() - pst->endDate()) > 0.001){
                of << " +"<< st->beginDate() - pst->endDate();
            }
            of<< "\""<< supp_ed <<" ];"<<endl;
        }
    }
    of<<'}'<<endl;
    of.close();
    system("sh -c 'dot task_time_graph.dot -Tsvg -o task_time_graph.svg && eog task_time_graph.svg' &");
}

void Task::removeIdles(){
    unsigned int i=0;
    while (i < _sub_tasks.size()){
        if(getSubTask(i)->type()==SubTask::IDLE){
            this->eraseSubTask(getSubTask(i));
        }else{
            i++;
        }
    }
}

void Task::fillWithIdle(map<MultiHandOverAgent*,MultiAgentCell*> & state_begin){
    double eps=0.0001;
    vector<MultiHandOverAgent*> all_agents;
    std::multimap<MultiHandOverAgent*,SubTask*> sub_tasks_by_agent;
    double start=numeric_limits<double>::infinity(), end=0;

    for(map<MultiHandOverAgent*,MultiAgentCell*>::iterator it=state_begin.begin();it!=state_begin.end();++it){
        all_agents.push_back(it->first);
    }
    for(unsigned int i=0;i<_sub_tasks.size();i++){
        sub_tasks_by_agent.insert(make_pair(getSubTask(i)->agent(), getSubTask(i)));
        start=min(start,getSubTask(i)->beginDate());
        end=max(end,getSubTask(i)->endDate());
    }

    for(unsigned int i=0;i<all_agents.size();i++){
        double s=start,e=end;
        vector<SubTask*> ordered_subtk;
        for(multimap<MultiHandOverAgent*,SubTask*>::iterator it=sub_tasks_by_agent.lower_bound(all_agents[i]);
            it!=sub_tasks_by_agent.upper_bound(all_agents[i]);it++){
            ordered_subtk.push_back(it->second);
        }
        std::sort(ordered_subtk.begin(),ordered_subtk.end(), SubTask::compareLessBeginDatePtr);
        SubTask* prev_tk(0);
        for(vector<SubTask*>::iterator it=ordered_subtk.begin();it!=ordered_subtk.end();++it){

            SubTask* tk=*it;
            e=tk->beginDate();
            if(std::abs(e - s) > eps && s < e){
                //doesnt follow directly
                SubTask *ntk=new SubTask(all_agents[i],SubTask::IDLE,s,e);
                if(prev_tk){
                    ntk->addWaypoint(prev_tk->waypoints2d().back());
                    ntk->addPrevious(prev_tk);
                    prev_tk->addNext(ntk);
                }else{
                    // the unique wp is == to the first of the following task [and the last of the previous]
                    ntk->addWaypoint(tk->waypoints2d().front());
                }
                ntk->addNext(tk);
                tk->addPrevious(ntk);
                ntk->spreadIdlePositionToNext();
                if(prev_tk){
                    double dist_test=(prev_tk->waypoints2d().back()-tk->waypoints2d().front()).norm();
                    //assert that prev and next coincide (idle does not involve navigation...!)
                    assert(dist_test<eps);
                }
                _sub_tasks.push_back(ntk);

            }
            //update start and end dates for next sub task
            s=tk->endDate();
            //e=end;
            prev_tk=tk;
        }
        if(s!=end){
            SubTask *ntk=new SubTask(all_agents[i],SubTask::IDLE,s,end);
            // the unique wp is == to the first of the following task [and the last of the previous]
            if(ordered_subtk.size()){
                SubTask *tk=ordered_subtk.back();
                ntk->addWaypoint(tk->waypoints2d().back());
                ntk->addPrevious(tk);
            }else{
                ntk->addWaypoint(state_begin[all_agents[i]]->getCenter());
            }
            _sub_tasks.push_back(ntk);
        }
    }
}

const std::vector<MultiHandOverAgent *> &Task::getAgents(){
    return _agents;
}

void Task::addSubTask(SubTask *st,bool no_sort){
    if(st){
        //add the agent to the list of agents
        if(find(_agents.begin(),_agents.end(),st->agent()) == _agents.end()){
            _agents.push_back(st->agent());
        }
        //add the task and sort by starting date
        _sub_tasks.push_back(st);
        if(!no_sort){
            sort(_sub_tasks.begin(),_sub_tasks.end(),SubTask::compareLessBeginDatePtr);
        }

    }
}
void Task::eraseSubTask(SubTask *st){
    vector<SubTask*>::iterator found_it;//=find(_sub_tasks.begin(),_sub_tasks.end(),st);
    for(vector<SubTask*>::iterator it=_sub_tasks.begin();it!=_sub_tasks.end();it++){
        if(*it==st){
            found_it=it;
        }else{
            //delete in prev/next lists
            SubTask* ost=*it;
            vector<SubTask*>::iterator iio=find(ost->getNext().begin(),ost->getNext().end(),st);
            if(iio!=ost->getNext().end()){
                ost->getNext().erase(iio);
            }
            iio=find(ost->getPrev().begin(),ost->getPrev().end(),st);
            if(iio!=ost->getPrev().end()){
                ost->getPrev().erase(iio);
            }
        }
    }
    if(found_it!=_sub_tasks.end()){
        delete *found_it;
        _sub_tasks.erase(found_it);
    }
}

Task::state_iterator Task::begin(){
    return state_iterator(this);
}

// /////// TASK STATE ITERATOR
Task::state_iterator::state_iterator(Task* task):
    _task(task),_is_end_iterator(0)
{
    _time=numeric_limits<double>::infinity();
    if(_task){
        for(unsigned int i=0;i<_task->getSubTasks().size();i++){
            MultiHandOverAgent* ag=_task->getSubTask(i)->agent();
            if(_sub_task.find(ag)==_sub_task.end()){
                //this agent as no first task yet
                _sub_task[ag]=_task->getSubTask(i);
                _state_current[ag]=_task->getSubTask(i)->waypoints2d().front();
                _time=min(_time,_task->getSubTask(i)->beginDate());
            }
        }
    }
}

double Task::state_iterator::date(){return _time ;}
bool Task::state_iterator::isEnd(){return _is_end_iterator;}

map<MHO::MultiHandOverAgent*,Eigen::Vector2d> & Task::state_iterator::state(){
    return _state_current;
}

SubTask* Task::state_iterator::getSubTask(MultiHandOverAgent *agent){
    return _sub_task[agent];
}

bool Task::state_iterator::advance(){
    //compute next time and state
    double time_proposal(numeric_limits<double>::infinity());
    for(unsigned int i=0;i<_task->getSubTasks().size();i++){
        MHO::SubTask *stk=_task->getSubTask(i);
        if(_time>=stk->beginDate()){
            time_proposal=min(_next[stk].second,time_proposal);
        }else{
            time_proposal=min(stk->beginDate(), time_proposal);
        }
        assert(time_proposal>=_time);
    }
    //stop condition
    if(time_proposal==numeric_limits<double>::infinity()){
        _is_end_iterator=true;
        return false;
    }

    //the time proposal is ok, update state
    _time=time_proposal;

    for(unsigned int i=0;i<_task->getSubTasks().size();i++){
        MHO::SubTask *stk=_task->getSubTask(i);
        //if at this date the task has started
        //if(stk->isNav() && time>=stk->beginDate()){
        if(_time>=stk->beginDate()){
            if(_next[stk].first==0){
                _next[stk].second=stk->beginDate();
            }
            //if the next wp should be reached at this time
            if(_time >= _next[stk].second){
                _sub_task[stk->agent()]=stk;
                //update current trak
                double next_time(0);
                _trak[stk]=_next[stk];

                Eigen::Vector2d curr_wp=stk->waypoints2d()[_trak[stk].first];

                //update state
                if(stk->type() != SubTask::IDLE){
                    _state_current[stk->agent()]=curr_wp;
                }

                //update next
                if(static_cast<unsigned int>(_trak[stk].first)+1 < stk->waypoints2d().size()){
                    int next_ind;//index of the next wp
                    next_ind=_trak[stk].first+1;
                    //next_ind=min(_trak[stk].first+1,(int)stk->waypoints2d().size()-1);
                    if(stk->type()==MHO::SubTask::NAV_SYNC){
                        next_time = stk->waypointDates()[next_ind];
                    }else{
                        Eigen::Vector2d next_wp=stk->waypoints2d()[next_ind];
                        Eigen::Vector2d diff=next_wp - curr_wp;
                        double dist = diff.norm();
                        next_time = _trak[stk].second + dist / stk->agent()->speed;
                    }
                    _next[stk]=make_pair(next_ind,next_time);
                }else if (_time>=stk->endDate()){
                    //reached end of this task
                    _next[stk]=make_pair(stk->waypoints2d().size(),numeric_limits<double>::infinity());
                }else{
                    // probably a task that does not involve navigation,
                    //(as last waypoint reached but not end date)
                    // next wp is still the current one, with date being
                    // the end date of the sub task
                    //if (stk->type() != SubTask::IDLE){
                        _next[stk]=make_pair(stk->waypoints2d().size()-1,stk->endDate());
                    //}//else : do not change anything
                }
            }else if(_next[stk].second==numeric_limits<double>::infinity()){
                if(_sub_task[stk->agent()]==stk){
                    //reset the sub task, as this one is done
                    //_sub_task[stk->agent()]=0;
                }
            }
            assert(_next[stk].second>=_time);
            assert(_next[stk].second>=_trak[stk].second);
            assert(_next[stk].second>=stk->beginDate());
        }
    }
    return true;
}

// ////////////////////////////////////////////////////
// ///////////     MHP HAND OVER OBJECT   /////////////
// ////////////////////////////////////////////////////

MultiHandOverMhpHO::MultiHandOverMhpHO(MultiHandOverSolution *sol, unsigned int step){
    if(!sol || step>=sol->actions.size()){
        std::string e(__FUNCTION__);
        e+="bad combinaison of solution/step in constructor call";
        throw std::invalid_argument(e);
    }
    MultiHandOverSolution::Action &act=sol->actions[step];
    //giver
    this->g_name=act.agent(0)->robot->getName();
    this->g_arm=ARM_RIGHT;
    confPtr_t conf=act.frontier()->confOf(act.agent(0)->robot);
    if (conf){
    //arm conf
    for(unsigned int i=0;i<7;i++){
        this->g_arm_conf.push_back(conf->at(16+i));
    }
    //torso
    this->g_torso=conf->at(12);
    //pose
    for(unsigned int i=0;i<2;i++){
        this->g_pose[i]=conf->at(6+i);
    }
    g_pose[2]=conf->at(11);//theta
    }else{
        cout<<"MHO WARNING: no conf for robot giver "<<g_name<<endl;
    }

    //receiver
    this->r_name=act.agent(1)->robot->getName();
    conf=act.frontier()->confOf(act.agent(1)->robot);
    if(conf){
        for(unsigned int i=0;i<2;i++){
            this->r_pose[i]=conf->at(6+i);
        }
        r_pose[2]=conf->at(11);//theta
        r_torso=conf->at(12);
        // cout << "receiver torso="<<r_torso<<endl
        //     << "receiver arm conf = ";
        for(unsigned int i=0;i<7;i++){
            this->r_arm_conf.push_back(conf->at(16+i));
            // cout<<r_arm_conf.back()<<", ";
        }
    }else{
        cout<<"MHO WARNING: no conf for robot receiver "<<r_name<<endl;
    }
    // cout<<endl;

    g_start_time=act.startDate(0);
    r_start_time=act.startDate(1);

}

// ////////////////////////////////////////////////////
// ///////////       HAND OVER OBJECT     /////////////
// ////////////////////////////////////////////////////

MultiHandOverHO::MultiHandOverHO(MultiHandOverSolution::Action *action, MultiHandOver* mho):
    _a1(action->agent(0)),_a2(action->agent(1)),
    _c1(action->frontier()->cellOf(_a1->index)),
    _c2(action->frontier()->cellOf(_a2->index)),
    _d(-1),
    _mho(mho)
{}

MultiHandOverHO::MultiHandOverHO(MultiHandOverAgent *a1, MultiHandOverAgent *a2, MultiAgentCell *c1, MultiAgentCell *c2, MultiHandOver* mho):
    _a1(a1),_a2(a2),_c1(c1),_c2(c2),_d(-1),
    _mho(mho)
{}
MultiHandOverHO::MultiHandOverHO(MultiHandOverAgent *a1, MultiHandOverAgent *a2, double d, MultiHandOver *mho):
    _a1(a1),_a2(a2),_c1(0),_c2(0),_d(d),
    _mho(mho)
{}

double MultiHandOverHO::computeCost(){
    double cost;
    if (robot(0)->getHriAgent()->is_human != robot(1)->getHriAgent()->is_human){
        //one is human, other is robot
        cost=humanRobotDistanceRelatedCost();
        this->setTime(3.5);//TODO handover time is dummy constant
    }else if (robot(0)->getHriAgent()->is_human && robot(1)->getHriAgent()->is_human){
        //both humans
        //TODO handover cost for r-r and h-h
        cost=0;
        this->setTime(2);//TODO handover time is dummy constant
    }else{
        //both robots
        //TODO handover cost for r-r and h-h
        cost=0;
        this->setTime(5);//TODO handover time is dummy constant
    }
    this->setTime(_mho->getTimeHO(_a1->index,_a2->index));
    setCost(cost);
    return cost;
}

void MultiHandOverHO::getOptimalCost(MultiHandOverAgent *a1, MultiHandOverAgent *a2,double &cost,double &duration,MultiHandOver* mho){
    cost=0;
    duration=mho->getTimeHO(a1->index,a2->index);
    //if(a1->is_human && a2->is_human){
        //duration=5;
    //}else{
        //duration=10;
    //}
}

double MultiHandOverHO::humanRobotDistanceRelatedCost(){
    double cost,dist,penetrationRatio,safe_dist,comfort_dist,comfortRatio,extensionDist,reach;
    //TODO dist_min unused
    if(_d<0){
        dist=this->cell(0)->computeDist(this->cell(1));
    }else{
        dist=_d;
    }
    if (agent(0)->is_human != agent(1)->is_human){
        //one is human, other is robot
        if(agent(0)->is_human){
            //dist_min=hri_robot_min_distance(robot(1)->getHriAgent(),robot(0)->getHriAgent());
            safe_dist=agent(0)->safe_dist;
            comfort_dist=agent(0)->handover_comfort_dist;
            extensionDist=dist-agent(1)->reach();
            reach=agent(0)->reach();
        }else{
            //dist_min=hri_robot_min_distance(robot(0)->getHriAgent(),robot(1)->getHriAgent());
            safe_dist=agent(1)->safe_dist;
            comfort_dist=agent(1)->handover_comfort_dist;
            extensionDist=dist-agent(0)->reach();
            reach=agent(1)->reach();
        }
        penetrationRatio = 1 - dist/safe_dist;
        if(penetrationRatio > 0 ){
            cost=pow( penetrationRatio,2);
        }else{cost=0;}
        //cost related to comfort
        //comfortRatio looks like:
        // ^ cost
        //1++           ++++
        // | +        +
        // |  +     +
        // |   +  +
        //0+----+--------+> dist
        // 0    c        e
        // c=comfortDist, e=extensionDist
        //is then elevated to square
        //TODO extension distance cost is false
        if(extensionDist<comfort_dist){
            comfortRatio = 1-extensionDist/comfort_dist;
        }else{
            comfortRatio = (extensionDist - comfort_dist) /
                    (reach - comfort_dist);
        }
        //assert(comfortRatio<1.01);
        if (comfortRatio>1){comfortRatio=1;} // impossible, dist>extensionDist means no HO possible
        cost+=pow(comfortRatio,2);
        cost/=2;

    }else{
        cost=-1;
    }
    this->setCost(cost);
    return cost;
}

//getters/setters
MultiHandOverAgent* MultiHandOverHO::agent(unsigned int i){return (i ? _a2 : _a1);}
Robot* MultiHandOverHO::robot(unsigned int i){return this->agent(i)->robot;}
MultiAgentCell* MultiHandOverHO::cell(unsigned int i){return (i ? _c2 : _c1);}
//FrontierPtr MultiHandOverHO::frontier(){return _frontier;}
double MultiHandOverHO::getCost() const{return this->cost;}
double MultiHandOverHO::getTime() const{return this->time;}
void MultiHandOverHO::setCost(double cost){this->cost=cost;}
void MultiHandOverHO::setTime(double time_sec){this->time=time_sec;}




// ////////////////////////////////////////////////////
// ///////////       HAND OVER FINDER     /////////////
// ////////////////////////////////////////////////////

HandOverFinder::HandOverFinder(MultiHandOver *mho, MultiAgentGrid *grid, unsigned int agent1, unsigned int agent2,
                               MultiAgentCell *cell1_O, MultiAgentCell *cell2_O,
                               MultiAgentCell *cell1_T, MultiAgentCell *cell2_T) :
    _grid(grid),_mho(mho),_agent1(agent1),_agent2(agent2),
    _cell1_ori(cell1_O), _cell2_ori(cell2_O),_cell1_dest(cell1_T),_cell2_dest(cell2_T)
{

}


FrontierPtr HandOverFinder::findFrom2Points(){
    vector<MultiAgentCell*> traj=_grid->computeOtherTrajectory(_cell1_ori,_cell2_ori,_agent1);
    if(traj.size()){
        double total_dist=_cell2_ori->getTempDist();
        double dist1=0;
        double prev;
        double cost=numeric_limits<double>::infinity();
        double K=0.9,offset(total_dist/2);
        while(std::fabs(offset)>0.001){
            prev=cost;
            cost = _mho->navigationRelatedCost(dist1,_agent1)+_mho->navigationRelatedCost(total_dist-dist1,_agent2);
            if(cost>=prev){
                offset=-offset;
            }
            offset=offset*K;
            dist1+=offset;
            dist1=max(min(dist1,total_dist),0.0);
            //cout<<"cost:"<<cost<<endl;
        }
        dist1-=offset;
        //cout <<"optimal distance is "<<dist1<< " over "<<total_dist<<endl;
    }
    FrontierPtr p;
    return p;

}

// ////////////////////////////////////////////////////
// ///////////       LAZY WEIGHTED A*     /////////////
// ////////////////////////////////////////////////////
MultiHandOverHeuristic::MultiHandOverHeuristic(MultiHandOver *mho, MHO::AbstractSolutionTools *s_tools):
    _mho(mho), _tools(s_tools)
{
    this->doTheGraph();
    rev_dijk.setGraph(&this->_graph);
    _times.assign(3,0);
}

double MultiHandOverHeuristic::compute(MultiAgentCell* c1,unsigned int a1,MultiAgentCell* c2,unsigned int a2){
    Chronometer chrono;
    std::pair<double,std::deque<mho::NodeInterface*> > dijk_res;
    //std::pair<double,std::deque<mho::NodeInterface*> > dijk_res;
    vector<unsigned int> agents;
    chrono.reset();
    if(rev_dijk.getTarget()!=_agents_nodes[a2]){
        rev_dijk.compute(_agents_nodes[a2]);
    }
    dijk_res=rev_dijk.getPathDeque(_agents_nodes[a1]);
    _times[0]+=chrono.getMilliSec();
    chrono.reset();
    double h=0,d(numeric_limits<double>::infinity());
    double dist=c1->computeDist(c2);
    for(unsigned int i=0;i<dijk_res.second.size();++i){
        agents.push_back(dijk_res.second[i]->getId());
        if(i>0){
            h+=_tools->handOverCostEstimation(agents[i-1],agents[i]);
            //the HO makes the economy of the distance related.
            //remove best case (max) distance from the overall distance
            dist=max(0.0,dist-_tools->handOverMaxDist(agents[i-1],agents[i]));
        }
    }
    _times[1]+=chrono.getMilliSec();
    chrono.reset();
    for(unsigned int i=0;i<agents.size();i++){
        d=min(d,_tools->distanceCost(dist,agents[i]));
    }
    _times[2]+=chrono.getMilliSec();
    h+= d;
    assert(h!=numeric_limits<double>::infinity());
    return h;
}


void MultiHandOverHeuristic::doTheGraph(){
    _graph.clear();
    _agents_nodes.clear();
    for(unsigned int i=0;i<_mho->getAgents().size();++i){
        mho::Node *n=new mho::Node(i);
        _agents_nodes.push_back(dynamic_cast<mho::Node*>(_graph.addNode(n)));
        delete n;
    }

    double tf,uf,df,hf;//time, use, distance, hri cost factors
    _mho->getCostsFactors(df,tf,uf,hf);
    for(unsigned int i=0;i<_agents_nodes.size();++i){
        for(unsigned int j=0;j<i;++j){
            if(_mho->handOverPossible(i,j)){
                //the cost of the handover is computed considering
                //a unity distance (d=1)
                double cost,c1,c2;
                //c= uf * (d* use_cost/speed) + tf * ( d/speed)
                c1=_tools->distanceCost(1,i);
                c2=_tools->distanceCost(1,j);
                //the cost is the minimum of the 2 distance related cost
                //as in a simple (best) case the cheaper will make all the distance
                cost=min(c1,c2) + _tools->handOverCostEstimation(i,j);
                _graph.connectNodes(_agents_nodes[i],_agents_nodes[j],cost);
            }
        }
    }
}

MultiHandOverHeuristicClosestFrontier::MultiHandOverHeuristicClosestFrontier(MultiHandOver *mho, AbstractSolutionTools *s_tools):
    MultiHandOverHeuristic(mho,s_tools)
{ }

double MultiHandOverHeuristicClosestFrontier::compute(MultiAgentCell *c1, unsigned int a1, MultiAgentCell *c2, unsigned int a2){
    double h=0;
    //cout << "H: ";
    if(a1 == a2){
        //cout << a1;
        //trivial
        h=this->getSolutionTools()->distanceCost(c1->computeDist(c2),a1);
    }else{
        std::pair<double,std::deque<mho::NodeInterface*> > dijk_res;
        vector<unsigned int> agents;
        if(getRevDijkstra().getTarget()!=getAgentNode(a2)){
            getRevDijkstra().compute(getAgentNode(a2));
        }
        dijk_res=getRevDijkstra().getPathDeque(getAgentNode(a1));
        for(unsigned int i=0;i<dijk_res.second.size();++i){
            agents.push_back(dijk_res.second[i]->getId());
            if(i>0){
                //cout <<dijk_res.second[i-1]->getId() << "-"<<dijk_res.second[i]->getId()<<", ";
                h+=getSolutionTools()->handOverCostEstimation(agents[i-1],agents[i]);

                MultiAgentCell *c1i(0),*c1g(0),*c2i(0),*c2g(0);
                // find the best frontier for that HO to minimize the distance ran by the 2 agents
                c1g = getMho()->getAgentTargetCell(agents[i-1]);//goes to target after
                c2i = getMho()->getAgentInitCell(agents[i]);//comes from init
                //others are not known (other HO), except:
                if(i==1){
                    //first HO, the origin of A1 is C1
                    c1i= c1;
                }
                if(i==dijk_res.second.size()-1){
                    //last HO, last agent goes to c2
                    c2g=c2;
                }
                //add the cost related to the optimal distances agents have to run.
                h+= findMinimalFrontier(c1i,c1g,agents[i-1],c2i,c2g,agents[i]);
            }
        }
    }
    //cout << endl;
    assert(h!=numeric_limits<double>::infinity());
    return h;
}

Eigen::Vector2d MultiHandOverHeuristicClosestFrontier::findMinimalFrontier2pt(Eigen::Vector2d c1, double f1, Eigen::Vector2d c2, double f2){
    Eigen::Vector2d g;

    if ( f1 < f2 ){
        // A1 cheaper, go to C2
        g = c2;
    }else if (f1 > f2 ){
        //A2 cheaper, go to C1
        g = c1;
    }else{
        // in the middle
        g = c1 + c2;
        g=g/2;
    }
    return g;
}

Eigen::Vector2d MultiHandOverHeuristicClosestFrontier::findMinimalFrontier3pt(Eigen::Vector2d ci,Eigen::Vector2d cg, double f1, Eigen::Vector2d c2, double f2){
    Eigen::Vector2d g;
    if (f1 < f2){
        // A1 is cheaper, g is c2
        g = c2;
    }else if (f2 < f1){
        //A2 is cheaper, g is the point of [c1i, c1g] the closest to c2i or c2g
        Eigen::Vector2d base= - ci + cg;
        Eigen::Vector2d seg2 = - ci + c2;
        double dot=base.dot(seg2);
        if (dot <= 0){
            // the point is ci
            g=ci;
        }else if (dot / base.norm() < base.norm() ){
            // the point is between ci and cg
            g = ci + dot / base.norm() * base;
        }else{
            //the closest point is cg
            g = cg;
        }
    }else{ // if (f1 == f2){
        g = this->findMinimalFrontier4pt(ci,cg,f1,c2,c2,f2/2);
    }
    return g;

}

Eigen::Vector2d MultiHandOverHeuristicClosestFrontier::findMinimalFrontier4pt(Vector2d c1i, Vector2d c1g, double f1, Vector2d c2i, Vector2d c2g, double f2){
    Eigen::Vector2d g;
    g = f1 * (c1i + c1g) + f2 * (c2i + c2g );
    g = g / ( f1 + f2 ) ;
    return g;
}

double MultiHandOverHeuristicClosestFrontier::findMinimalFrontier(
        MultiAgentCell *c1i, MultiAgentCell *c1g, unsigned int a1,
        MultiAgentCell *c2i, MultiAgentCell *c2g, unsigned int a2)
{
    double cost= numeric_limits<double>::infinity();
    if(! getMho()->getPlanGrid()->getAgentStructAt(a1)->getMetaAgent()->
            isFrontierComputed(
                getMho()->getPlanGrid()->getAgentStructAt(a2)->getMetaAgent()))
    {
        //we search the point G such that the sum of distances run by each agent is minimal.
        // this is the centroid of every start and target weighted with the cost factor of the agent
        Eigen::Vector2d g(0,0);
        double f1 = getSolutionTools()->distanceCost(1,a1),
                f2 = getSolutionTools()->distanceCost(1,a2);
        double d1(0),d2(0);

        if(c1g && c1i && c2i && c2g ){
            if(c1i != c1g && c2i != c2g){
     // CAS 1   // 4 points all differents: get the centroid
                g = this->findMinimalFrontier4pt(c1i->getCenter(), c1g->getCenter(), f1, c2i->getCenter(), c2g->getCenter(), f2);
            }else if (c1i != c1g ){
     // CAS 4a  // 2 different for A1 and 2 identical for A2
                // multiply f2 by 2 to take this into account
                g = this->findMinimalFrontier3pt(c1i->getCenter(),c1g->getCenter(),f1,(c2i ? c2i->getCenter() : c2g->getCenter()),f2 * 2);
            }else if (c2i != c2g ){
     // CAS 4b  // 2 different for A2 and 2 identical for A1
                // multiply f1 by 2 to take this into account
                g = this->findMinimalFrontier3pt(c2i->getCenter(),c2g->getCenter(),f2,(c1i ? c1i->getCenter() : c1g->getCenter()),f1 * 2);
            } else {
     // CAS 3   //identical init and goal for both
                assert (c1i == c1g && c2i == c2g );
                g = this->findMinimalFrontier2pt(c1i->getCenter(), f1 * 2, c2i->getCenter(),f2 * 2);
            }
        }else if( c1g && c1i && c1g != c1i && (c2g || c2i)){
     //CAS 2a// two points for A1 and one for A2
            g = this->findMinimalFrontier3pt(c1i->getCenter(),c1g->getCenter(),f1,(c2i ? c2i->getCenter() : c2g->getCenter()),f2);
        }else if( c2g && c2i && c2g != c2i && (c1g || c1i)){
     //CAS 2b// two points for A2 and one for A1
            g = this->findMinimalFrontier3pt(c2i->getCenter(),c2g->getCenter(),f2,(c1i ? c1i->getCenter() : c1g->getCenter()),f1);
        }else if ( (c1g || c1i) && (c2g || c2i) ){
     //CAS 0 // one point each
            g = this->findMinimalFrontier2pt( ( c1i ? c1i->getCenter() : c1g->getCenter() ), f1,
                                              ( c2i ? c2i->getCenter() : c2g->getCenter() ), f2 );
        }else if (c1i || c1g){
            assert ( !c2i && !c2g);
     //CAS 5a //only for A1
            g = ( c1i ? c1i->getCenter() : c1g->getCenter() );

        }else if (c2i || c2g){
            assert ( !c1i && !c1g);
     //CAS 5b //only for A2
            g = ( c2i ? c2i->getCenter() : c2g->getCenter() );
        }else{
            assert (false);
        }

        if(c1i){ d1 += (c1i->getCenter() - g ).norm(); }
        if(c1g){ d1 += (c1g->getCenter() - g ).norm(); }
        if(c2i){ d2 += (c2i->getCenter() - g ).norm(); }
        if(c2g){ d2 += (c2g->getCenter() - g ).norm(); }

        cost = d1*f1 +d2 *f2;
        if ( abs(cost - getSolutionTools()->distanceCost(d1,a1) - getSolutionTools()->distanceCost(d2,a2)) >= 1e-6){
            cerr << cost << " vs "<< getSolutionTools()->distanceCost(d1,a1) + getSolutionTools()->distanceCost(d2,a2)<< endl;
            assert(abs(cost - getSolutionTools()->distanceCost(d1,a1) - getSolutionTools()->distanceCost(d2,a2)) < 1e-6);
        }

    }else{
        vector<MultiAgentCell*> cells=getMho()->getPlanGrid()->getAgentStructAt(a1)->frontiers(getMho()->getPlanGrid()->getAgentStructAt(a2));
        for (unsigned int i =0;i<cells.size();i++){
            MultiAgentCell* cA= cells[i];
            double d1=(c1i ? c1i->computeDist(cA) : 0 ) + (c1g ? c1g->computeDist(cA) : 0 );
            double c1 = getSolutionTools()->distanceCost(d1,a1);
            vector<MetaFrontierPtr> & ftrs = cA->getMetaFrontier(a1,a2);
            for (unsigned int fi=0;fi < ftrs.size() ; fi++){
                MetaFrontierPtr ftr=ftrs[fi];
                MultiAgentCell *cB = ftr->cellOf(cA->getMetaFromGrid(a2));
                double d2=(c2i ? c2i->computeDist(cB) : 0 ) + (c2g ? c2g->computeDist(cB) : 0 );
                double c = c1 + getSolutionTools()->distanceCost(d2,a2);
                cost = min (cost , c);
                if (c == 0){
                    return cost;
                }
            }
        }
    }
    return cost;
}

LazyWeightedAstarNode::LazyWeightedAstarNode(MultiAgentCell* cell, unsigned int agent,MultiHandOverHeuristic *heuristic, ToolSet *tool_set):
    LWANodeBase(), _cell(cell), _agent(agent),
    _heuristic_class(heuristic),
    _toolset(tool_set)
{
    if(cell){
        _agents_in_parents = vector<bool>(cell->getAgentNumber(),false);
    }else{
        _agents_in_parents = vector<bool>(agent+1,false);
    }
    _agents_in_parents[agent]=true;
}

LazyWeightedAstarNode::LazyWeightedAstarNode(LazyWeightedAstarNode *other):
    LWANodeBase(other), _cell(other->cell()), _agent(other->agent()),
    _heuristic_class(other->_heuristic_class), _toolset(other->_toolset),
    _agents_in_parents(other->_agents_in_parents)
{}

LazyWeightedAstarNode::~LazyWeightedAstarNode(){ }

mho::LWANodeInterface * LazyWeightedAstarNode::clone(){
    mho::LWANodeInterface *clone = new LazyWeightedAstarNode(this);
    addClone(cast(clone));
    return clone;
}

vector<mho::LWANodeInterface*> LazyWeightedAstarNode::sons(){
    vector<mho::LWANodeInterface*> sons;
    vector<MultiAgentCell*> neigh;
    neigh=this->cell()->getNeighbors(this->agent());
    for(unsigned int i=0;i<neigh.size();++i){
        sons.push_back(new LazyWeightedAstarNode(neigh[i],this->agent(),_heuristic_class,_toolset));
        assert(cast(sons.back())->cell() != this->cell());
    }
    unsigned int n_agents=cell()->getAgentNumber();
    if(_agents_in_parents.size()<cell()->getAgentNumber()){
        _agents_in_parents.resize(cell()->getAgentNumber(),false);
    }
    for(unsigned int i=0;i<n_agents;++i){
        if(!_agents_in_parents[i] && _toolset->solutionTools->distanceCost(1,i) < numeric_limits<double>::infinity()){
            vector<FrontierPtr> frontiers = cell()->createAllRealFrontiersFor(agent(),i);
            //vector<FrontierPtr> &frontiers = this->cell()->getAgentAt(this->agent())->frontiers(this->cell()->getAgentAt(i));
            for(unsigned int f=0;f<frontiers.size();++f){
                sons.push_back(new LazyWeightedAstarNode(frontiers[f]->cellOf(i),i,_heuristic_class,_toolset));
                assert(cast(sons.back())->cell() != this->cell());
            }
            vector<MultiAgentCell*> neig_in_range=this->cell()->findNeighborsInRange(i,this->_toolset->getReach(this->agent(),i));
            double crit;
            if(neig_in_range.size() <= 10){crit=1.01;}//crit is more than one => add all
            else{
                crit=min(static_cast<double>(pow(neig_in_range.size(),-0.4) ), 20./neig_in_range.size());//TODO go to params
            }
            for(vector<MultiAgentCell*>::iterator it=neig_in_range.begin();it!=neig_in_range.end();++it){
                if(rand()/RAND_MAX < crit){
                    //double d1 = std::fabs((*it)->getDistance(this->agent())-this->cell()->getDistance(this->agent()))*1.5;
                    //double d2 = std::fabs((*it)->getDistance(i)-this->cell()->getDistance(i))*1.5;
                    //only add if one can travel between the 2 cells
                    //if(min(d1,d2) < this->_toolset->getReach(agent(),i)){
                    sons.push_back(new LazyWeightedAstarNode(*it,i,_heuristic_class,_toolset));
                    //}
                }
            }
        }
    }
    return sons;
}

double LazyWeightedAstarNode::edgeCostEstimationTo(LWANodeInterface *other){
    //compute the distances
    double c;
    if ( this->agent() == cast(other)->agent()){
        double d=this->cell()->computeDist(cast(other)->cell());
        c=this->_toolset->solutionTools->distanceCost(d,this->agent());
    }else{
        //c=_toolset->solutionTools->handOverCostEstimation(agent(),cast(other)->agent());
        c=_toolset->solutionTools->handOverCost(cell(),agent(),cast(other)->cell(),cast(other)->agent());
        c+=_toolset->solutionTools->handOverSideCosts(this->cell(),this->agent(),cast(other)->cell(),cast(other)->agent());
    }
    return c;
}

double LazyWeightedAstarNode::computeHeuristic(LWANodeInterface* goal){
    return _heuristic_class->compute(this->cell(),this->agent(),cast(goal)->cell(),cast(goal)->agent());
}

double LazyWeightedAstarNode::computeTrueCost(LWANodeInterface *other){
    //all distances, check hand over confs
    double c(0);
    if ( this->agent() == cast(other)->agent()){
        double d=this->cell()->computeDist(cast(other)->cell());
        c=this->_toolset->solutionTools->distanceCost(d,this->agent());
    }else{
        bool ok;
        ok=_toolset->checkHandOverFull(cast(other)->cell(),cast(other)->agent(),this->cell(),this->agent(),c);
        //TODO add go back and coming costs
        if(ok){
            c+=_toolset->solutionTools->handOverSideCosts(this->cell(),this->agent(),cast(other)->cell(),cast(other)->agent());
        }
    }
    return c;

}

bool LazyWeightedAstarNode::isSameConf(LWANodeInterface *other){
    return cast(other) && this->cell() == cast(other)->cell() && this->agent() == cast(other)->agent();
}

void LazyWeightedAstarNode::parent(LazyWeightedAstarNode *p){
    LWANodeBase::parent(p);
    this->_agents_in_parents=p->_agents_in_parents;
    if (p->_agents_in_parents.size() <= this->_agent){
        _agents_in_parents.resize(this->cell()->getAgentNumber(),false);
    }
    _agents_in_parents[_agent]=true;
}

// /////// DISPLAY OF LWA* //////////////////////////

DisplayLWAstarInGrid::DisplayLWAstarInGrid(MultiAgentGrid *grid,bool showtext,bool draw):
    _grid(grid),_marker_is_first_call(1),_showText(showtext),_draw(draw),_max_draw_f(-1)
{
}

void DisplayLWAstarInGrid::onOpen(mho::LWANodeInterface *open){
    MultiAgentCell* c=LazyWeightedAstarNode::cast(open)->cell();
    if(_marker_is_first_call){
        _marker_is_first_call=false;
        if(_showText){
            cout<<"LWA* starts"<<endl;
        }
    }
    if(c->getTempColor() == -1){
        if(open->isTrueCost() && open->g()< numeric_limits<double>::infinity()){
            c->setTempColor(open->g());
            _max_draw_f = max(_max_draw_f,open->g());
        }else if(open->isTrueCost()){
            c->setTempColor(-3);//invalid cost
        }else{
            //tmp cost
            c->setTempColor(-2);
        }
        _cells.push_back(c);
        if(this->_draw){
            _grid->setDrawExtraCells(_cells.begin(),_cells.end(),_max_draw_f);
            //g3d_draw_allwin_active();
        }
    }else{
        if(c->getTempColor()>=0 ){
            if(open->isTrueCost() && open->g() < numeric_limits<double>::infinity()){
                c->setTempColor(min(open->g(),c->getTempColor()));
            }
        }else if(open->isTrueCost()){
            c->setTempColor(open->g());
        }
    }
    count_open++;
}

void DisplayLWAstarInGrid::onClose(mho::LWANodeInterface *closed){
    MultiAgentCell* c=LazyWeightedAstarNode::cast(closed)->cell();
    if(c->getTempColor() >=0 ){
        c->setTempColor(min(c->getTempColor(),closed->g()));
    }else if( closed->g() < numeric_limits<double>::infinity()){
        c->setTempColor(closed->g());
    }else{
        c->setTempColor(-3);
    }
    if( closed->g() < numeric_limits<double>::infinity()){
        _max_draw_f = max(_max_draw_f,closed->g());
    }
    //c->setClosed();
    //if(c->getTempDist()<0){c->setTempDist(closed->f());}
    //else{
    //    c->setTempDist(min(c->getTempDist(),closed->f()));
    //}
    //if(this->_draw){
    //    _grid->setDrawExtraCells(_cells.begin(),_cells.end());
    //}
    _grid->setDrawExtraCells(_cells.begin(),_cells.end(),_max_draw_f);
    if(_draw){
        g3d_draw_allwin_active();
    }
    count_closed++;
    //c->resetExplorationStatus();
}

std::vector<std::pair<unsigned int,double> > DisplayLWAstarInGrid::getTimes(){
    return _times;
}

void DisplayLWAstarInGrid::onEnd(std::vector<pair<unsigned int, double> > times){
    _times=times;
    if(_showText){
        cout<<"LWA* end with "<< count_open <<" states explored and " << count_closed <<" states closed"<<endl;
        cout<< "times in LWA* search:"<<endl;
        int i=0;
        cout<< "\tfind similar closed:        "
            <<times[i].first<<'\t'<<times[i].second<<endl;
        i++;
        cout<< "\tnode expansion:             "
            <<times[i].first<<'\t'<<times[i].second<<endl;
        i++;
        cout<< "\testimation (included in 1): "
            <<times[i].first<<'\t'<<times[i].second<<endl;
        i++;
        cout<< "\theur comp (included in 1):  "
            <<times[i].first<<'\t'<<times[i].second<<endl;
        i++;
        cout<< "\ttrue cost comp and treat:   "
            <<times[i].first<<'\t'<<times[i].second<<endl;
        i++;
        cout<< "\ttrue cost comp (inc in 4):  "
            <<times[i].first<<'\t'<<times[i].second<<endl;
        i++;
        cout<< "\tclone (inc in 1):           "
            <<times[i].first<<'\t'<<times[i].second<<endl;
        i++;
        cout<< "\tinterf. in 1:               "
            <<times[i].first<<'\t'<<times[i].second<<endl;
        i++;
        cout<< "\thave cheaper clone (in 1):  "
            <<times[i].first<<'\t'<<times[i].second<<endl;
        i++;
        cout<< "\tget sons (in 1):            "
            <<times[i].first<<'\t'<<times[i].second<<endl;
    }
    for(unsigned int i=0;i<_cells.size();++i){
        //_cells[i]->resetExplorationStatus();
        _cells[i]->setTempColor(-1);
    }
    _grid->unsetDrawExtraCells();
}
void DisplayLWAstarInGrid::onFound(mho::LWANodeInterface *goal){
    LazyWeightedAstarNode *n=LazyWeightedAstarNode::cast(goal);
    while(n!=0){
        //n->cell()->setTempDist(-1);
        n->cell()->setTempColor(-4);
        n=LazyWeightedAstarNode::cast(n->parent());
    }
    if(_draw){
        g3d_draw_allwin_active();
    }
}
}
}
