#include "MultiPathPlanner.h"

using namespace std;
using namespace mho;
using namespace HRICS;

MultiPathPlanner::MultiPathPlanner():
    _epsilon(4),_start(0),_goal(0)
{
    _display=new MultiPathPlannerLWADisplayInterface();
    _lwa=new LazyWeightedAstar(_epsilon,_display);
}

MultiPathPlanner::~MultiPathPlanner(){
    delete _lwa;
    delete _display;
    _goal=0;_start=0;
    this->reset();
}

void MultiPathPlanner::reset(){
    this->_agents_indices.clear();
    this->_start_pos.clear();
    this->_goal_pos.clear();
    if(_goal) delete _goal;
    this->_goal=0;
    if(_start) delete _start;
    this->_start=0;
}

void MultiPathPlanner::setAgents(std::vector<unsigned int> agents_ids){
    this->reset();
    this->_agents_indices=agents_ids;
    this->_goal_pos.assign(agents_ids.size(),0);
    this->_start_pos.assign(agents_ids.size(),0);
}
void MultiPathPlanner::setGoal(std::vector<MultiAgentCell *> cells){
    if(this->_agents_indices.size()==cells.size()){
        this->_goal_pos=cells;
    }else{
        throw std::invalid_argument("MultiPathPlanner: There are not as much cells than agents...");
    }
}
void MultiPathPlanner::setStart(std::vector<MultiAgentCell *> cells){
    if(this->_agents_indices.size()==cells.size()){
        this->_start_pos=cells;
    }else{
        throw std::invalid_argument("MultiPathPlanner: There are not as much cells than agents...");
    }
}
void MultiPathPlanner::setStart(unsigned int index, MultiAgentCell *start){
    if(index<_agents_indices.size()){
        this->_start_pos[index]=start;
    }
}
void MultiPathPlanner::setGoal(unsigned int index, MultiAgentCell *goal){
    if(index<_agents_indices.size()){
        this->_goal_pos[index]=goal;
    }
}
void MultiPathPlanner::setStartByID(unsigned int agent_id, MultiAgentCell *start){
    std::vector<unsigned int>::iterator it=std::find(_agents_indices.begin(),_agents_indices.end(),agent_id);
    if(it!=_agents_indices.end()){
        _start_pos[it-_agents_indices.begin()]=start;
    }
}
void MultiPathPlanner::setGoalByID(unsigned int agent_id, MultiAgentCell *goal){
    std::vector<unsigned int>::iterator it=std::find(_agents_indices.begin(),_agents_indices.end(),agent_id);
    if(it!=_agents_indices.end()){
        _goal_pos[it-_agents_indices.begin()]=goal;
    }
}

void MultiPathPlanner::setSolutionTools(MHO::AbstractSolutionTools *sol_tools){
    this->_solution_tools=sol_tools;
}

bool MultiPathPlanner::compute(double time_limit_ms){
    bool ok=true;
    //start pos are all defined ?:
    ok=(find(_start_pos.begin(),_start_pos.end(),(MultiAgentCell*)0) == _start_pos.end());
    if(ok){
        // solution tools is defined ?
        ok=(_solution_tools != 0);
    }
    if(ok && _hierachy){
        std::deque <mho::LWANodeInterface* > prev_path;
        mho::LazyWeightedAstar * prev_lwa=0;
        for(unsigned int agent_i=0;agent_i<_agents_indices.size();agent_i++){
            //subvectors to adress only current agent and higher priority ones
            vector<unsigned int> a_id_aux(_agents_indices.begin(),_agents_indices.begin()+agent_i+1);
            vector<MultiAgentCell*> start_aux(_start_pos.begin(),_start_pos.begin()+agent_i+1),
                    goal_aux(_goal_pos.begin(),_goal_pos.begin()+agent_i+1);
            if(agent_i==0){
                _start=new MultiPathLWANode(a_id_aux,start_aux,this->_solution_tools);
                _goal=new MultiPathLWANode(a_id_aux,goal_aux,this->_solution_tools);
            }else{
                if(prev_lwa){delete prev_lwa;}
                prev_lwa=_lwa;//have to save to reuse prev_path elements (ptrs owned by _lwa)
                _lwa=new LazyWeightedAstar(_epsilon,_display);
                _start=new MultiPathLWANode(a_id_aux,start_aux,this->_solution_tools,&prev_path,prev_path.begin());
                _goal=new MultiPathLWANode(a_id_aux,goal_aux,this->_solution_tools,&prev_path,prev_path.begin());
                //for(unsigned int a_i_b=0;a_i_b <= agent_i;a_i_b++){
                //    _start->freeWainting(a_i_b,MultiPathLWANode::BEGIN);
                //    _goal->freeWainting(a_i_b,MultiPathLWANode::BEGIN);
                //}
            }
            _start->freeWainting(1,MultiPathLWANode::BEGIN);
            _goal->freeWainting(1,MultiPathLWANode::BEGIN);
            _goal->setStartGoal(_start,_goal);
            _start->setStartGoal(_start,_goal);
            ok=_lwa->compute(_start,_goal,time_limit_ms);
            if(ok){
                prev_path = getPath();
            }else{
                break;
            }
        }
        if(prev_lwa){delete prev_lwa;}
    }else if(ok){
        _start=new MultiPathLWANode(_agents_indices,_start_pos,this->_solution_tools);
        _goal=new MultiPathLWANode(_agents_indices,_goal_pos,this->_solution_tools);
        _start->setStartGoal(_start,_goal);
        _goal->setStartGoal(_start,_goal);
        _start->freeWainting(1,MultiPathLWANode::BEGIN);
        _goal->freeWainting(1,MultiPathLWANode::BEGIN);
        ok=_lwa->compute(_start,_goal,time_limit_ms);
    }
    return ok;
}

std::deque<LWANodeInterface*> MultiPathPlanner::getPath(){
    return (_lwa->getPath());
}

void MultiPathPlanner::setDisplay(bool showText, bool drawGrid, MultiAgentGrid *grid){
    this->_display->setShowText(showText);
    this->_display->setDraw(drawGrid,grid);
}

MultiPathLWANode::MultiPathLWANode(): LWANodeBase() {
}

MultiPathLWANode::MultiPathLWANode(vector<unsigned int> agents, std::vector<HRICS::MultiAgentCell *> positions,
                                   HRICS::MHO::AbstractSolutionTools *sol_tools ):
    LWANodeBase(),
    _agents_indices(agents),_positions(positions),
    _start(0),_goal(0),_duration_max(0),
    _solution_tools(sol_tools), _prev_path(0)
{
    _wait_is_free.assign(_agents_indices.size(),NEVER);
}

MultiPathLWANode::MultiPathLWANode(std::vector<unsigned int> agents, std::vector<MultiAgentCell *> positions,
                                   MHO::AbstractSolutionTools *sol_tools,
                                   std::deque<LWANodeInterface *> *prev_paths, deque<LWANodeInterface*>::iterator prev_path_current):
    LWANodeBase(),
    _agents_indices(agents),_positions(positions),
    _start(0),_goal(0),_duration_max(0),
    _solution_tools(sol_tools), _prev_path(prev_paths),_prev_path_current(prev_path_current)
{
    _wait_is_free.assign(_agents_indices.size(),NEVER);
}


MultiPathLWANode::MultiPathLWANode(MultiPathLWANode *other):
    LWANodeBase(other),
    _agents_indices(other->_agents_indices),_positions(other->_positions),
    _start(other->_start),_goal(other->_goal),_duration_max(other->_duration_max),
     _wait_is_free(other->_wait_is_free),
    _solution_tools(other->_solution_tools),
    _prev_path(other->_prev_path),
    _prev_path_current(other->_prev_path_current)
{
}

MultiPathLWANode::~MultiPathLWANode(){
}

LWANodeInterface* MultiPathLWANode::clone(){
    mho::LWANodeInterface *clone=new MultiPathLWANode(this);
    addClone(cast(clone));
    return clone;
}

vector<LWANodeInterface*> MultiPathLWANode::sons(){
    vector<LWANodeInterface*> sons;

    if(_prev_path){
        assert(_prev_path_current!=_prev_path->end());
        //2 "base" sons : already planned move to next step, or dont move (all do the same)
        vector<vector<MultiAgentCell*> > son_list_of_cell_cur(1,vector<MultiAgentCell*>(_agents_indices.size(),0));
        vector<vector<MultiAgentCell*> > son_list_of_cell_next(1,vector<MultiAgentCell*>(_agents_indices.size(),0));
        MultiPathLWANode *prev_path_current=dynamic_cast<MultiPathLWANode*>(*_prev_path_current);
        MultiPathLWANode *prev_path_next=0;
        if(_prev_path_current+1!=_prev_path->end()){
            prev_path_next=dynamic_cast<MultiPathLWANode*>(*(_prev_path_current+1));
            assert(prev_path_next);
        }
        assert(prev_path_current);

        vector<vector<MultiAgentCell* > > new_sons_list_of_cell_cur;
        vector<vector<MultiAgentCell* > > new_sons_list_of_cell_next;

        for(unsigned int prev_ag_i=0;prev_ag_i<prev_path_current->_agents_indices.size();prev_ag_i++){
            son_list_of_cell_cur.front()[prev_ag_i]=prev_path_current->_positions[prev_ag_i];
        }
        if(prev_path_next){
            for(unsigned int prev_ag_i=0;prev_ag_i<prev_path_current->_agents_indices.size();prev_ag_i++){
                son_list_of_cell_next.front()[prev_ag_i]=prev_path_next->_positions[prev_ag_i];
            }
        }else{
            son_list_of_cell_next.clear();
        }

        //then combinate with all sons from remaining agents
        for(unsigned int new_ag_i=prev_path_current->_agents_indices.size();new_ag_i<_agents_indices.size();new_ag_i++){
            vector<MultiAgentCell*> neighs=_positions[new_ag_i]->getNeighbors(_agents_indices[new_ag_i]);
            new_sons_list_of_cell_cur.clear();
            new_sons_list_of_cell_next.clear();
            //for each son found, for each neighbour of this agent,
            // append son U neighbour to the new list of sons
            //when done for all sons and neighbours, replace the son list by the new son list, and reiterate for next agent
            for(unsigned int son_i=0;son_i<son_list_of_cell_cur.size();son_i++){
                for(unsigned int neigh_i=0;neigh_i<neighs.size();neigh_i++){
                    //for both current state from already computed path, and next state in thAT path
                    new_sons_list_of_cell_cur.push_back(son_list_of_cell_cur[son_i]); // copy the list of cell
                    new_sons_list_of_cell_cur.back()[new_ag_i]=(neighs[neigh_i]);  // append a neighbour

                    if(prev_path_next){
                        new_sons_list_of_cell_next.push_back(son_list_of_cell_next[son_i]);
                        new_sons_list_of_cell_next.back()[new_ag_i]=(neighs[neigh_i]);
                    }
                }
            }
            son_list_of_cell_cur=new_sons_list_of_cell_cur;
            new_sons_list_of_cell_cur.clear();
            son_list_of_cell_next=new_sons_list_of_cell_next;
            new_sons_list_of_cell_next.clear();
        }
        for(unsigned int i=0;i<son_list_of_cell_cur.size();i++){
            assert(son_list_of_cell_cur[i].size()==_agents_indices.size());
            sons.push_back(new MultiPathLWANode(_agents_indices,son_list_of_cell_cur[i],_solution_tools,_prev_path,_prev_path_current));
            cast(sons.back())->setStartGoal(_start,_goal);
        }
        for(unsigned int i=0;i<son_list_of_cell_next.size();i++){
            assert(son_list_of_cell_next[i].size()==_agents_indices.size());
            sons.push_back(new MultiPathLWANode(_agents_indices,son_list_of_cell_next[i],_solution_tools,_prev_path,_prev_path_current+1));
            cast(sons.back())->setStartGoal(_start,_goal);
        }
    }else{
        vector<vector<MultiAgentCell*> > neigh_cell_lists;// list of neighbor cell for each agent
        vector<unsigned int> nb_of_comb(_agents_indices.size(),0);
        unsigned int nb_of_sons(1); // the number of sons, before they re all computed

        for(unsigned int i=0;i<_agents_indices.size();++i){
            unsigned int a=_agents_indices[i];
            MultiAgentCell* c=_positions[i];
            neigh_cell_lists.push_back(c->getNeighbors(a));
            neigh_cell_lists.back().push_back(c);
            nb_of_sons*=neigh_cell_lists[i].size();
            assert(neigh_cell_lists[i].size()>0);
        }


        nb_of_comb.back()=1;//neigh_cell_lists.back().size();
        for(int i=neigh_cell_lists.size()-2;i>=0;--i){
            nb_of_comb[i]=nb_of_comb[i+1]*neigh_cell_lists[i+1].size();
        }
        assert(nb_of_comb[0]*neigh_cell_lists[0].size()==nb_of_sons);

        sons.reserve(nb_of_sons);

        //fill the array of sons
        //one son is a list of cells
        //the array looks like follows:
        // A A A
        // A A B
        // A B A
        // A B B
        // A C A
        // A C B
        // A D A
        //...
        //assuming there are 3 agents, the third as 2 neigh cell, the 2nd as at least 4
        for(unsigned int i=0;i<nb_of_sons;i++){
            vector<MultiAgentCell*> son_list_of_cell(_agents_indices.size(),0);
            for(unsigned int ia=0;ia<_agents_indices.size();++ia)
            {
                unsigned int i_n=(i/nb_of_comb[ia]) % neigh_cell_lists[ia].size();//the index of the cell to add
                son_list_of_cell[ia]=neigh_cell_lists[ia][i_n];
            }

            sons.push_back(new MultiPathLWANode(_agents_indices,son_list_of_cell,_solution_tools));
            cast(sons.back())->setStartGoal(_start,_goal);
        }
    }

    return sons;
}

double MultiPathLWANode::edgeCostEstimationTo(LWANodeInterface *other){
    MultiPathLWANode *o=cast(other);
    double c=0;
    double t=0;
    for(unsigned int i=0;i<_agents_indices.size();++i){
        double d=0,time=0;
        if(cell(i) && o->cell(i)){
            d=this->cell(i)->computeDist(o->cell(i));
        }
        time=this->_solution_tools->distanceTime(d,this->agent_index(i));
        // first, the cost is for the distance, we remove the time component (use cost, busyness-related)
        if (_goal->cell(i)){
            //it is free for the ones that have no goal. We can't manage them
            c+=this->_solution_tools->distanceCost(d,this->agent_index(i))-this->_solution_tools->timeCost(time,this->agent_index(i));
        }
        t=max(t,this->_solution_tools->distanceTime(d,this->agent_index(i)));
    }

    //add the total time to each agent (the use cost of one agent is related to the total time
    //to execute this step, ie the time of the slower agent, not the time this agent
    //take to run its distance
    for(unsigned int i=0;i<_agents_indices.size();++i){
        if ( (this->freeWainting(i)==BEGIN && this->cell(i)==this->_start->cell(i))
             ||  (this->freeWainting(i)==END && this->cell(i)==this->_goal->cell(i))){
            continue;
        }else{
            c+=this->_solution_tools->timeCost(t,this->agent_index(i));
        }
    }

    _duration_max=t;
    return c;
}

double MultiPathLWANode::computeHeuristic(LWANodeInterface *goal){
    MultiPathLWANode *n_goal=cast(goal);
    double h=0;
    for(unsigned int i=0;i<_agents_indices.size();++i){
        double d=0;
        if(cell(i) && n_goal->cell(i)){
            d=this->cell(i)->computeDist(n_goal->cell(i));
        }
        h+=this->_solution_tools->distanceCost(d,this->agent_index(i));
    }
    return h;
}

double MultiPathLWANode::computeTrueCost(LWANodeInterface *other){
    //MultiPathLWANode *o=cast(other);
    double c=0;
    bool col=false;
    //check for colisions
    for (unsigned int i=0;i<_agents_indices.size();i++){
        for (unsigned int j=0;j<i;j++){
            if(cell(i)!=0 && cell(j)!=0 && cell(i)->checkIsInCollision(agent_index(i),cell(j),agent_index(j))){
                col=true;
                break;
            }
        }
        if(col){
            break;
        }
    }
    if(!col){
        c=this->edgeCostEstimationTo(other);
    }else{
        c=numeric_limits<double>::infinity();
    }
    return c;
}

bool MultiPathLWANode::isSameConf(LWANodeInterface *other){
    MultiPathLWANode *o=cast(other);
    bool is_identical=true;
    for (unsigned int i=0;i<_agents_indices.size();++i){
        if((cell(i) != o->cell(i)) && cell(i)!=0 && o->cell(i)!=0){
            //if one is not defined, its ok (0==whatever, 0!=nothing)
            is_identical=false;
            break;
        }
    }
    //assert(this ==other || other->isSameConf(other) && this->isSameConf(this));
    return is_identical;
}



void MultiPathPlannerLWADisplayInterface::onOpen(mho::LWANodeInterface* open){
    //if(_showText){
    //    std::cout<<"o-- f:"<<open->f()<<"; g:"<<open->g()<<"; h:"<<open->h()<<std::endl;
    //}
    _open_c++;
}
void MultiPathPlannerLWADisplayInterface::onClose(mho::LWANodeInterface* closed){
    //if(_showText){
    //  std::cout<<"c-- f:"<<closed->f()<<"; g:"<<closed->g()<<"; h:"<<closed->h()<<std::endl;
    //}
    if(_draw && _grid && closed->g() < numeric_limits<double>::infinity()){
        MultiPathLWANode* n=MultiPathLWANode::cast(closed);
        for(unsigned int i=0;i<_cells_for_draw.size();++i){
            _cells_for_draw[i]->resetExplorationStatus();
        }
        _cells_for_draw.clear();
            MultiPathLWANode *current=n;
            while(current){
                for(unsigned int i=0;n->agent_index(i)>=0;i++){
                    MultiAgentCell* c=current->cell(i);
                    c->setClosed();
                    c->setTempDist(i-1);
                    _cells_for_draw.push_back(c);
                }
                current=MultiPathLWANode::cast(current->parent());
            }
        _grid->setDrawExtraCells(_cells_for_draw.begin(),_cells_for_draw.end(),1);
        g3d_draw_allwin_active();

    }
    _closed_c++;
}
void MultiPathPlannerLWADisplayInterface::onFound(mho::LWANodeInterface* goal){
    if(_showText){
        std::cout<<"found!"<<std::endl;
    }
    if(_draw && _grid){
        MultiPathLWANode* n=MultiPathLWANode::cast(goal);
        for(unsigned int i=0;i<_cells_for_draw.size();++i){
            _cells_for_draw[i]->resetExplorationStatus();
        }
        _cells_for_draw.clear();
        for(unsigned int i=0;n->agent_index(i)>=0;i++){
            n->cell(i)->setClosed();
            n->cell(i)->setTempDist(n->agent_index(i)*10);
            _cells_for_draw.push_back(n->cell(i));
        }
        _grid->setDrawExtraCells(_cells_for_draw.begin(),_cells_for_draw.end(),1);
        g3d_draw_allwin_active();

    }
}
void MultiPathPlannerLWADisplayInterface::onEnd(std::vector<std::pair<unsigned int, double> > times){
    if(_showText){
        std::cout<<"LWA finished, o="<<_open_c<<" ; c="<<_closed_c<<endl;
    }
    for(unsigned int i=0;i<_cells_for_draw.size();++i){
        _cells_for_draw[i]->resetExplorationStatus();
    }
    _grid->unsetDrawExtraCells();
    _cells_for_draw.clear();
}

// ////////////////////////////////////////////////////
// ////////////       MULTI PATH CHECK        /////////
// ////////////////////////////////////////////////////

MultiPathCheck::MultiPathCheck(MultiAgentGrid *grid, MHO::Task *task, std::map<MHO::MultiHandOverAgent *,
                               MultiAgentCell *> start_state):
    _tk(task), _start_state(start_state), _grid(grid)
{
}

bool MultiPathCheck::check(){
    //std::map<MHO::MultiHandOverAgent*,MultiAgentCell*> state = _start_state;
    //std::map<MHO::MultiHandOverAgent*,MHO::SubTask*> sub_task;
    //std::map<MHO::SubTask*,pair<int, double> > trak; //saves the current/reached waypoint index of each SubTask and the time associated
    //std::map<MHO::SubTask*,pair<int, double> > next; //saves the next waypoint index of each SubTask and the time associated
    _colliding_stk.clear();
    //double time=0;
    bool ok=true;
    //bool end=false;
    //MHO::Task::state_iterator st_it=_tk->begin();
    //while (!end){
    for(MHO::Task::state_iterator st_it=_tk->begin();!st_it.isEnd();st_it.advance()){
        //for(map<MHO::MultiHandOverAgent*,Eigen::Vector2d>::iterator st_it_st=st_it.state().begin();
        //    st_it_st!= st_it.state().end();st_it_st++){
        //    assert(st_it_st->second == state[st_it_st->first]->getCenter());
        //}
        //check this state
        for(map<MHO::MultiHandOverAgent*,Eigen::Vector2d>::iterator st_it_st=st_it.state().begin();
            st_it_st!= st_it.state().end();st_it_st++){
        //for(map<MHO::MultiHandOverAgent*,MultiAgentCell*>::iterator it=state.begin();it!=state.end();it++){
            for(map<MHO::MultiHandOverAgent*,Eigen::Vector2d>::iterator st_it_st2=st_it.state().begin();
                st_it_st2!= st_it_st;st_it_st2++){
            //for(map<MHO::MultiHandOverAgent*,MultiAgentCell*>::iterator it2=state.begin();it2!=it;it2++){
                bool col=false;
                MultiAgentCell* c1=dynamic_cast<MultiAgentCell*>(_grid->getCell(st_it_st->second)),
                        *c2=dynamic_cast<MultiAgentCell*>(_grid->getCell(st_it_st2->second));
                MHO::MultiHandOverAgent *a1=st_it_st->first,*a2=st_it_st2->first;
                col=c1->checkIsInCollision(a1->index,c2,a2->index);
                if(col){
                    Configuration q=*a1->robot->getInitialPosition();
                    q[6+0]=c1->getCenter()[0];
                    q[6+1]=c1->getCenter()[1];
                    a1->robot->setAndUpdate(q);
                    Configuration q2=*a2->robot->getInitialPosition();
                    q2[6+0]=c2->getCenter()[0];
                    q2[6+1]=c2->getCenter()[1];
                    a2->robot->setAndUpdate(q2);
                    g3d_draw_allwin_active();
                    if(std::find(_colliding_stk.begin(),_colliding_stk.end(),make_pair(st_it.getSubTask(a1),st_it.getSubTask(a2)))==_colliding_stk.end()){
                        ok=false;
                        _colliding_stk.push_back(make_pair(st_it.getSubTask(a1),st_it.getSubTask(a2)));
                    }
                }
            }
        }

        // //compute next time and state
        // double time_proposal(numeric_limits<double>::infinity());
        // for(unsigned int i=0;i<_tk->getSubTasks().size();i++){
        //     MHO::SubTask *stk=_tk->getSubTask(i);
        //     //if(stk->isNav()){
        //         if(time>=stk->beginDate()){
        //             time_proposal=min(next[stk].second,time_proposal);
        //         }else{
        //             time_proposal=min(stk->beginDate(), time_proposal);
        //         }
        //     //}
        // }

        //st_it.advance();
        //stop condition
        // if(st_it.isEnd()){
        //     break;
        // }
        // if(time_proposal==numeric_limits<double>::infinity()){
        //     assert(st_it.isEnd());
        //     end=true;
        //     break;
        // }

        //the time proposal is ok, update state
        //time=time_proposal;
        //time=st_it.date();

        // for(unsigned int i=0;i<_tk->getSubTasks().size();i++){
        //     MHO::SubTask *stk=_tk->getSubTask(i);
        //     //if at this date the task has started
        //     //if(stk->isNav() && time>=stk->beginDate()){
        //     if(time>=stk->beginDate()){
        //         //if the next wp should be reached at this time
        //         if(time >= next[stk].second){
        //             sub_task[stk->agent()]=stk;
        //             //update current trak
        //             double next_time(0);
        //             trak[stk]=next[stk];

        //             Eigen::Vector2d curr_wp=stk->waypoints2d()[trak[stk].first];

        //             //update state
        //             state[stk->agent()]=dynamic_cast<MultiAgentCell*>(_grid->getCell(curr_wp));

        //             //update next
        //             if(static_cast<unsigned int>(trak[stk].first)+1 < stk->waypoints2d().size()){
        //                 int next_ind;//index of the next wp
        //                 next_ind=trak[stk].first+1;
        //                 next_ind=min(trak[stk].first+1,(int)stk->waypoints2d().size()-1);
        //                 Eigen::Vector2d next_wp=stk->waypoints2d()[next_ind];
        //                 Eigen::Vector2d diff=next_wp - curr_wp;
        //                 double dist = diff.norm();
        //                 next_time = trak[stk].second + dist / stk->agent()->speed;
        //                 next[stk]=make_pair(next_ind,next_time);
        //             }else if (time>=stk->endDate()){
        //                 //reached end of this task
        //                 next[stk]=make_pair(stk->waypoints2d().size(),numeric_limits<double>::infinity());
        //             }else{
        //                 // probably a task that does not involve navigation,
        //                 // next wp is still the current one, with date being
        //                 // the end date of the sub task
        //                 next[stk]=make_pair(stk->waypoints2d().size()-1,stk->endDate());
        //             }
        //         }else if(next[stk].second==numeric_limits<double>::infinity()){
        //             if(sub_task[stk->agent()]==stk){
        //                 //reset the sub task, as this one is done
        //                 sub_task[stk->agent()]=0;
        //             }
        //         }
        //     }
        // }

    }
    return ok;
}
















